import abc
from pathlib import Path

from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.environment import Label


class MLMethod(metaclass=abc.ABCMeta):
    """
    Base class for different machine learning methods, defining which functions should be implemented. These public functions are the only ones that
    will be used outside the method, during training, assessment or while making predictions. Most often the methods will be classifiers (binary or
    multi-class) that should learn some label on either immune repertoires (sets of receptor sequences), receptors (paired sequences) or receptor
    sequences (lists of amino acids).

    Here we refer to machine learning methods (algorithms) as a method that, given a set of examples and corresponding labels, constructs a model
    (such as logistic regression), whereas we define the model to be already fit to data using the learning method (algorithm), such as logistic
    regression with specific coefficients.

    The functions of this class provide a standard set of ML functions: fitting the model (with or without cross-validation) and making predictions
    (either class predictions or class probabilities if possible). Other functions provide for various utilities, such as storing and loading the
    model, checking if it was fit already, retrieving coefficients for user-friendly output etc.

    Note that when providing class probabilities the classes should have a specific (constant) order, and in case of binary classification, they
    should be ordered so that the negative class comes first and the positive one comes second. For this handling classes, see
    py:`immuneML.ml_methods.util.Util.Util.make_class_mapping` method that will automatically create class mapping for classification.

    """
    def __init__(self):
        self.name = None
        self.label = None

    @abc.abstractmethod
    def fit(self, encoded_data: EncodedData, label: Label, optimization_metric: str, cores_for_training: int = 2):
        """
        The fit function fits the parameters of the machine learning model.

        Arguments:

            encoded_data (EncodedData): an instance of EncodedData class which includes encoded examples (repertoires, receptors or sequences), their
                labels, names of the features and other additional information. Most often, only examples and labels will be used. Examples are either a
                dense numpy matrix or a sparse matrix, where columns correspond to features and rows correspond to examples. There are a few encodings
                which make multidimensional outputs that do not follow this pattern, but they are tailored to specific ML methods which require such input
                (for instance, one hot encoding and ReceptorCNN method).

            label (Label): the label for which the classifier will be created. immuneML also supports multi-label classification, but it is
                handled outside MLMethod class by creating an MLMethod instance for each label. This means that each MLMethod should handle only one label.

            optimization_metric (str): the name of the optimization metric to be used to select the best model during cross-validation; when used with
                TrainMLModel instruction which is almost exclusively the case when the immuneML is run from the specification, this maps to the
                optimization metric in the instruction.

            cores_for_training (int): if parallelization is available in the MLMethod (and the availability depends on the specific classifier), this
                is the number of processes that will be creating when fitting the model to speed up the computation.


        Returns:

            it doesn't return anything, but fits the model parameters instead

        """
        pass

    @abc.abstractmethod
    def predict(self, encoded_data: EncodedData, label: Label):
        """
        The predict function predicts the class for the given label across examples provided in encoded data.

        Arguments:

            encoded_data (EncodedData): an instance of EncodedData class which includes encoded examples (repertoires, receptors or sequences), their
                labels, names of the features and other additional information. Most often, only examples and labels will be used. Examples are either a
                dense numpy matrix or a sparse matrix, where columns correspond to features and rows correspond to examples. There are a few encodings
                which make multidimensional outputs that do not follow this pattern, but they are tailored to specific ML methods which require such input
                (for instance, one hot encoding and ReceptorCNN method).

            label: (Label): the label for which the classifier will be created. immuneML also supports multi-label classification, but it is
                handled outside MLMethod class by creating an MLMethod instance for each label. This means that each MLMethod should handle only one label.

        Returns:

            a dictionary where the key is the label_name and the value is a list of class predictions (one prediction per example):
            e.g., {label_name: [class1, class2, class2, class1]}

        """
        pass

    @abc.abstractmethod
    def fit_by_cross_validation(self, encoded_data: EncodedData, label: Label, optimization_metric: str, number_of_splits: int = 5, cores_for_training: int = -1):
        """
        The fit_by_cross_validation function should implement finding the best model hyperparameters through cross-validation. In immuneML,
        preprocessing, encoding and ML hyperparameters can be optimized by using nested cross-validation (see TrainMLModelInstruction for more
        details). This function is in that setting the third level of nested cross-validation as it can optimize only over the model hyperparameters.
        It represents an alternative to optimizing the model hyperparameters in the TrainMLModelInstruction. Which one should be used depends on the
        use-case and specific models: models based on scikit-learn implementations come with this option by default (see SklearnMethod class), while
        custom classifiers typically do not implement this and just call fit() function and throw a warning instead.

        Arguments:

            encoded_data (EncodedData): an instance of EncodedData class which includes encoded examples (repertoires, receptors or sequences), their
                labels, names of the features and other additional information. Most often, only examples and labels will be used. Examples are either a
                dense numpy matrix or a sparse matrix, where columns correspond to features and rows correspond to examples. There are a few encodings
                which make multidimensional outputs that do not follow this pattern, but they are tailored to specific ML methods which require such input
                (for instance, one hot encoding and ReceptorCNN method).

            label (Label): the label for which the classifier will be created. immuneML also supports multi-label classification, but it is
                handled outside MLMethod class by creating an MLMethod instance for each label. This means that each MLMethod should handle only one label.

            optimization_metric (str): the name of the optimization metric to be used to select the best model during cross-validation; when used with
                TrainMLModel instruction which is almost exclusively the case when the immuneML is run from the specification, this maps to the
                optimization metric in the instruction.

            number_of_splits (int): number of splits for the cross-validation to be performed for selection the best hyperparameters of the ML model;
                note that if this is used in combination with nested cross-validation in TrainMLModel instruction, it can result in very few examples in
                each split depending on the orginal dataset size and the nested cross-validation setup.

            cores_for_training (int): number of processes to be used during the cross-validation for model selection


        Returns:

            it doesn't return anything, but fits the model parameters instead

        """
        pass

    @abc.abstractmethod
    def store(self, path: Path, feature_names: list = None, details_path: Path = None):
        """
        The store function stores the object on which it is called so that it can be imported later using load function. It typically uses pickle,
        yaml or similar modules to store the information. It can store one or multiple files.

        Arguments:

            path (Path): path to folder where to store the model

            feature_names (list): list of feature names in the encoded data; this can be stored as well to make it easier to map linear models to
                specific features as provided by the encoded (e.g., in case of logistic regression, this feature list defines what coefficients refer to)

            details_path (Path): path to folder where to store the details of the model. The details can be there to better understand the model but
                are not mandatory and are typically not loaded with the model afterwards. This is user-friendly file that can be examined manually by the
                user. It does not have to be created or can be created at the same folder as the path parameters points to. In practice, when used with
                TrainMLModel instruction, this parameter will either be None or have the same value as path parameter.

        Returns:

            it does not have a return value

        """
        pass

    @abc.abstractmethod
    def load(self, path: Path):
        """
        The load function can load the model given the folder where the same class of the model was previously stored using the store function.
        It reads in the parameters of the model and sets the values to the object attributes so that the model can be reused. For instance, this is
        used in MLApplication instruction when the previously trained model is applied on a new dataset.

        Arguments:

            path (Path): path to the folder where the model was stored using store() function

        Returns:

            it does not have a return value, but sets the attribute values of the object instead

        """
        pass

    @abc.abstractmethod
    def check_if_exists(self, path: Path) -> bool:
        """
        The check_if_exists function checks if there is a stored model on the given path. Might be useful in the future for implementing checkpoints.
        See SklearnMethod for example usage.

        Arguments:

            path (Path): path to folder where it should be checked if the model was stored previously

        Returns:

            True/False: whether the model was stored previously on the given path or not

        """
        pass

    @abc.abstractmethod
    def get_params(self):
        """Returns the model parameters in a readable yaml-friendly way (consisting of lists, dictionaries and strings)."""
        pass

    @abc.abstractmethod
    def predict_proba(self, encoded_data: EncodedData, Label: Label):
        """
        The predict_proba function predicts class probabilities for the given label if the model supports probabilistic output. If not, it should
        raise a warning and return predicted classes without probabilities.

        The function will return a nested dictionary. The key(s) of the outer dictionary represent the label name(s),
        and the keys of the inner dictionary the class names of the respective label.
        The utility function py:mod:`immuneML.ml_methods.util.Util.Util.make_binary_class_mapping` may be used to
        handle mapping of class names to an internal representation for binary classification.

        Arguments:

            encoded_data (EncodedData): an object of EncodedData class where the examples attribute should be used to make predictions. examples
            attribute includes encoded examples in matrix format (numpy 2D array or a sparse matrix depending on the encoding). EncodedData object
            provided here can include labels (in model assessment scenario) or not (when predicting the class probabilities on new data which has not
            been labels), so the labels attribute of the EncodedData object should NOT be used in this function, even if it is set.

            label (Label): the label for which the prediction should be made. It can be used to check if it matches the label that the
            model has been trained for and if not, an exception should be thrown. It is often an AssertionError as this can be checked before any
            prediction is made, but could also be a RuntimeError. It both cases, it should include a user-friendly message.

        Returns:

            a nested dictionary where the outer keys represent label names, inner keys represent class names for the respective
            label, and innermost values are 1D numpy arrays with class probabilities.
            For example for instance for label CMV where the class can be either True or False and there are
            3 examples to predict the class probabilities for: {CMV: {True: [0.2, 0.55, 0.98], False: [0.8, 0.45, 0.02]}}

        """
        pass

    @abc.abstractmethod
    def get_label_name(self) -> str:
        """Returns the name of the label for which the model was fitted."""
        pass

    @abc.abstractmethod
    def get_package_info(self) -> str:
        """
        Returns the package and version used for implementing the ML method if an external package was used or immuneML version if it is custom
        implementation. See py:mod:`immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod` and py:mod:`immuneML.ml_methods.ProbabilisticBinaryClassifier.ProbabilisticBinaryClassifier`
        for examples of both.
        """
        pass

    @abc.abstractmethod
    def get_feature_names(self) -> list:
        """
        Returns the list of feature names (a list of strings) if available where the feature names were provided by the encoder in the
        EncodedData object.
        """
        pass

    @abc.abstractmethod
    def can_predict_proba(self) -> bool:
        """
        Returns whether the ML model can be used to predict class probabilities or class assignment only.
        """
        return False

    @abc.abstractmethod
    def get_class_mapping(self) -> dict:
        """Returns a dictionary containing the mapping between label values and values internally used in the classifier"""
        pass

    @abc.abstractmethod
    def get_compatible_encoders(self):
        pass

    def check_encoder_compatibility(self, encoder):
        """Checks whether the given encoder is compatible with this ML method, and throws an error if it is not."""
        is_valid = False

        for encoder_class in self.get_compatible_encoders():
            if issubclass(encoder.__class__, encoder_class):
                is_valid = True
                break

        if not is_valid:
            raise ValueError(f"{encoder.__class__.__name__} is not compatible with ML Method {self.__class__.__name__}. "
                             f"Please use one of the following encoders instead: {', '.join([enc_class.__name__ for enc_class in self.get_compatible_encoders()])}")

    def get_classes(self):
        """The get_classes function returns a list of classes for which the method was trained."""
        return self.label.values

    def get_positive_class(self):
        """The get_positive_class function returns the positive class for which the method was trained."""
        return self.label.positive_class



import abc
from pathlib import Path

import numpy as np

from immuneML.environment.SequenceType import SequenceType


class GenerativeModel:
    OUTPUT_COLUMNS = []

    def __init__(self, chain):
        self.chain = chain

    @abc.abstractmethod
    def fit(self, data, path: Path = None):
        pass

    @abc.abstractmethod
    def is_same(self, model) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def generate_sequences(self, count: int, seed: int, path: Path, sequence_type: SequenceType, compute_p_gen: bool):
        pass

    @abc.abstractmethod
    def compute_p_gens(self, sequences, sequence_type: SequenceType) -> np.ndarray:
        pass

    @abc.abstractmethod
    def compute_p_gen(self, sequence: dict, sequence_type: SequenceType) -> float:
        pass

    @abc.abstractmethod
    def can_compute_p_gens(self) -> bool:
        pass

    @abc.abstractmethod
    def can_generate_from_skewed_gene_models(self) -> bool:
        pass

    @abc.abstractmethod
    def generate_from_skewed_gene_models(self, v_genes: list, j_genes: list, seed: int, path: Path,
                                         sequence_type: SequenceType, batch_size: int,
                                         compute_p_gen: bool):
        pass

    @abc.abstractmethod
    def save_model(self, path: Path) -> Path:
        pass

    @classmethod
    @abc.abstractmethod
    def load_model(cls, path: Path):
        pass



import abc
import copy
from itertools import chain
from typing import List
from uuid import uuid4

from immuneML.data_model.DatasetItem import DatasetItem
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.data_model.receptor.receptor_sequence.SequenceMetadata import SequenceMetadata
from immuneML.util.NumpyHelper import NumpyHelper


class Receptor(DatasetItem):
    FIELDS = {}

    def __init__(self, identifier: str = None, metadata: dict = None):
        self.metadata = metadata
        self.identifier = identifier if identifier is not None else uuid4().hex

    @abc.abstractmethod
    def get_chains(self):
        pass

    @classmethod
    def create_from_record(cls, **kwargs):

        chains = cls().get_chains()

        chain1_record = {key.replace(f'{chains[0]}_', ''): val for key, val in kwargs.items() if
                         key.startswith(chains[0])}

        chain2_record = {key.replace(f'{chains[1]}_', ''): val for key, val in kwargs.items() if
                         key.startswith(chains[1])}

        assert chain1_record['cell_id'] == chain2_record['cell_id'], (chain1_record['cell_id'], chain2_record['cell_id'])
        if 'identifier' in chain1_record:
            assert chain1_record['identifier'] == chain2_record['identifier'], (chain1_record['identifier'], chain2_record['identifier'])
            identifier = kwargs[f'{chains[0]}_identifier']
        else:
            identifier = kwargs[f'{chains[0]}_cell_id']

        return cls(**{chains[0]: ReceptorSequence.create_from_record(**chain1_record),
                      chains[1]: ReceptorSequence.create_from_record(**chain2_record),
                      'identifier': identifier,
                      'metadata': {key.replace(f'{chains[0]}_', '').replace(f'{chains[1]}_', ''): val
                                   for key, val in kwargs.items()
                                   if key.replace(f'{chains[0]}_', '').replace(f'{chains[1]}_', '') not in ReceptorSequence.FIELDS.keys() and key.replace(f'{chains[0]}_', '').replace(f'{chains[1]}_', '') not in vars(SequenceMetadata()).keys()}})

    @classmethod
    @abc.abstractmethod
    def get_record_names(cls):
        pass

    def clone(self):
        copied_element = copy.deepcopy(self)
        copied_element.identifier = str(uuid4().hex)
        return copied_element

    def get_chain(self, chain: str):
        return getattr(self, chain)

    def get_record(self):
        chains = self.get_chains()
        record = self.get_chain(chains[0]).get_record() + self.get_chain(chains[1]).get_record() \
                 + [NumpyHelper.get_numpy_representation(getattr(self, name)) for name in self.FIELDS if
                    name not in chains]

        return record

    def get_id(self):
        return self.identifier

    def get_attribute(self, name: str):
        if hasattr(self, name):
            return getattr(self, name)
        if self.metadata is not None and name in self.metadata.keys():
            return self.metadata[name].to_string() if hasattr(self.metadata[name], 'to_string') else self.metadata[name]
        else:
            return [self.get_chain(ch).get_attribute(name) for ch in self.get_chains()]

    def get_all_attribute_names(self) -> List[str]:
        names = list(self.metadata.keys()) if self.metadata is not None else [] + ['cell_id']
        names += ['identifier']
        names += list(set(chain.from_iterable([self.get_chain(ch).get_all_attribute_names() for ch in self.get_chains()])))
        return names

    def __repr__(self):
        return self.__class__.__name__ + f"(identifier={self.identifier})"



import pandas as pd

from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.ImportHelper import ImportHelper


class AdaptiveImportHelper:

    @staticmethod
    def preprocess_dataframe(dataframe: pd.DataFrame, params: DatasetImportParams):
        if "frame_type" in dataframe.columns:
            dataframe.loc[:, "frame_type"] = dataframe.frame_type.str.upper()

            frame_type_list = ImportHelper.prepare_frame_type_list(params)
            dataframe = dataframe[dataframe["frame_type"].isin(frame_type_list)]

        dataframe.loc[:, "region_type"] = params.region_type.name

        if params.region_type == RegionType.IMGT_CDR3:
            if "sequence" in dataframe.columns:
                dataframe.loc[:, 'sequence'] = [y[(84 - 3 * len(x)): 78] if x is not None else None for x, y in
                                                 zip(dataframe['sequence_aa'], dataframe['sequence'])]
            dataframe.loc[:, 'sequence_aa'] = dataframe["sequence_aa"].str[1:-1]
        elif "sequences" in dataframe.columns:
            dataframe.loc[:, 'sequence'] = [y[(81 - 3 * len(x)): 81] if x is not None else None for x, y in
                                             zip(dataframe['sequence_aa'], dataframe['sequence'])]

        dataframe = AdaptiveImportHelper.parse_adaptive_germline_to_imgt(dataframe, params.organism)
        ImportHelper.load_chains(dataframe)
        ImportHelper.drop_empty_sequences(dataframe, params.import_empty_aa_sequences, params.import_empty_nt_sequences)
        ImportHelper.drop_illegal_character_sequences(dataframe, params.import_illegal_characters, params.import_with_stop_codon)

        return dataframe

    @staticmethod
    def parse_adaptive_germline_to_imgt(dataframe, organism):
        gene_name_replacement = pd.read_csv(
            EnvironmentSettings.root_path / "immuneML/IO/dataset_import/conversion/imgt_adaptive_conversion.csv")
        gene_name_replacement = gene_name_replacement[gene_name_replacement.Species == organism]
        gene_name_replacement = dict(zip(gene_name_replacement.Adaptive, gene_name_replacement.IMGT))

        # remove C and extra 0 from gene name but not from allele (e.g., TCRBV03-01*01 -> TRBV3-1*01) to follow IMGT naming
        germline_value_replacement = {**{"TCRB": "TRB", "TCRA": "TRA"},
                                      **{f"-0{i}": f"-{str(i)}" for i in range(10)},
                                      **{f"J0": "J", "V0": "V"}}

        return AdaptiveImportHelper.parse_germline(dataframe, gene_name_replacement, germline_value_replacement)

    @staticmethod
    def parse_germline(dataframe: pd.DataFrame, gene_name_replacement: dict, germline_value_replacement: dict):
        for gene in ["v", "j"]:
            for part in ['gene', 'allele', 'call', 'family']:
                if f"{gene}_{part}" in dataframe.columns:
                    dataframe.loc[:, f"{gene}_{part}"] = dataframe[f"{gene}_{part}"].replace(gene_name_replacement, regex=True)
                    if part != 'family':
                        dataframe.loc[:, f"{gene}_{part}"] = dataframe[f"{gene}_{part}"].replace(germline_value_replacement, regex=True)

            if f"{gene}_call" in dataframe.columns and f"{gene}_allele" in dataframe.columns:
                rows_to_add_allele = ~dataframe[f'{gene}_call'].str.contains("\*") & dataframe[f"{gene}_allele"].astype(str).str.contains("[0-9]{2}")
                dataframe.loc[rows_to_add_allele, f"{gene}_call"] = \
                    dataframe.loc[rows_to_add_allele, lambda df: [f"{gene}_call", f"{gene}_allele"]].agg('*'.join, axis=1)

        return dataframe



from typing import List

from bionumpy import AminoAcidEncoding, DNAEncoding
from bionumpy.bnpdataclass import bnpdataclass

from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.environment.SequenceType import SequenceType


@bnpdataclass
class BackgroundSequences:
    sequence_aa: AminoAcidEncoding
    sequence: DNAEncoding
    v_call: str
    j_call: str
    region_type: str
    frame_type: str
    p_gen: float
    from_default_model: int
    duplicate_count: int
    chain: str

    def get_sequence(self, sequence_type: SequenceType = SequenceType.AMINO_ACID):
        if sequence_type == SequenceType.AMINO_ACID:
            return self.sequence_aa
        else:
            return self.sequence

    @classmethod
    def build_from_receptor_sequences(cls, sequences: List[ReceptorSequence]):
        return BackgroundSequences(sequence_aa=[s.sequence_aa for s in sequences],
                                   sequence=[s.sequence for s in sequences],
                                   v_call=[s.metadata.v_call if s.metadata else '' for s in sequences],
                                   j_call=[s.metadata.j_call if s.metadata else '' for s in sequences],
                                   region_type=[s.metadata.region_type.name if s.metadata else '' for s in sequences],
                                   frame_type=[s.metadata.frame_type.name if s.metadata else '' for s in sequences],
                                   p_gen=[-1. for _ in sequences], from_default_model=[1 for _ in sequences],
                                   duplicate_count=[s.metadata.duplicate_count for s in sequences])



import warnings
from collections import Counter
from pathlib import Path
from typing import Union

import pandas as pd
import plotly.express as px

from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class SequenceLengthDistribution(DataReport):
    """
    Generates a histogram of the lengths of the sequences in a repertoire or sequence dataset.

    Specification arguments:

    - sequence_type (str): whether to check the length of amino acid or nucleotide sequences; default value is 'amino_acid'

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_sld_report:
            SequenceLengthDistribution:
                sequence_type: amino_acid

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_sequence_type(kwargs)

        return SequenceLengthDistribution(**{**kwargs, 'sequence_type': SequenceType[kwargs['sequence_type'].upper()]})

    def __init__(self, dataset: Union[RepertoireDataset, SequenceDataset] = None, batch_size: int = 1, result_path: Path = None, number_of_processes: int = 1,
                 sequence_type: SequenceType = SequenceType.AMINO_ACID, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.batch_size = batch_size
        self.sequence_type = sequence_type

    def check_prerequisites(self):
        if isinstance(self.dataset, RepertoireDataset) or isinstance(self.dataset, SequenceDataset):
            return True
        else:
            warnings.warn("SequenceLengthDistribution: report can be generated only from repertoire and sequence datasets. Skipping this report...")
            return False

    def _generate(self) -> ReportResult:
        sequence_lengths = self._get_sequence_lengths()
        PathBuilder.build(self.result_path)
        df = pd.DataFrame({"counts": list(sequence_lengths.values()), 'sequence_lengths': list(sequence_lengths.keys())})
        df.to_csv(self.result_path / 'sequence_length_distribution.csv', index=False)

        report_output_fig = self._safe_plot(df=df, output_written=False)
        output_figures = None if report_output_fig is None else [report_output_fig]
        return ReportResult(name=self.name,
                            info="A histogram of the lengths of the sequences in a dataset.",
                            output_figures=output_figures, output_tables=[ReportOutput(self.result_path / 'sequence_length_distribution.csv',
                                                                                       'lengths of sequences in the dataset')])

    def _get_sequence_lengths(self) -> Counter:

        if isinstance(self.dataset, RepertoireDataset):
            sequence_lengths = Counter()

            for repertoire in self.dataset.get_data(self.batch_size):
                seq_lengths = self._count_in_repertoire(repertoire)
                sequence_lengths += seq_lengths
        else:
            sequence_lengths = {}
            for sequence in self.dataset.get_data():
                l = len(sequence.get_sequence(self.sequence_type))
                if l in sequence_lengths:
                    sequence_lengths[l] += 1
                else:
                    sequence_lengths[l] = 1

        return sequence_lengths

    def _count_in_repertoire(self, repertoire: Repertoire) -> Counter:
        c = Counter([len(sequence.get_sequence(self.sequence_type)) for sequence in repertoire.sequences])
        return c

    def _plot(self, df: pd.DataFrame) -> ReportOutput:

        figure = px.bar(df, x="sequence_lengths", y="counts")
        figure.update_layout(xaxis=dict(tickmode='array', tickvals=df["sequence_lengths"]), yaxis=dict(tickmode='array', tickvals=df["counts"]),
                             template="plotly_white")
        figure.update_traces(marker_color=px.colors.diverging.Tealrose[0])
        PathBuilder.build(self.result_path)

        file_path = self.result_path / "sequence_length_distribution.html"
        figure.write_html(str(file_path))
        return ReportOutput(path=file_path, name="Sequence length distribution plot")




# quality: gold
import copy
import logging
import uuid
from pathlib import Path

import pandas as pd

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.environment.Constants import Constants
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class RepertoireDataset(Dataset):

    @classmethod
    def build_from_objects(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()), ['repertoires', 'path'], RepertoireDataset.__name__, RepertoireDataset.__name__)
        ParameterValidator.assert_all_type_and_value(kwargs['repertoires'], Repertoire, RepertoireDataset.__name__, 'repertoires')

        assert len(kwargs['repertoires']) > 0, "Cannot to construct a repertoire dataset without repertories."

        metadata_df = pd.DataFrame.from_records([{**rep.metadata, **{'filename': rep.data_filename}} for rep in kwargs['repertoires']])

        if 'field_list' in metadata_df.columns:
            metadata_df.drop(columns=['field_list'], inplace=True)

        metadata_path = PathBuilder.build(kwargs['path']) / 'metadata.csv'
        metadata_df.to_csv(metadata_path, index=False)

        logging.info(f"Made new repertoire dataset at {kwargs['path']} with metadata at {metadata_path} "
                     f"with {len(kwargs['repertoires'])} repertoires.")

        dataset = RepertoireDataset(repertoires=kwargs['repertoires'], metadata_file=metadata_path)
        label_names = list(dataset.get_label_names(refresh=True))
        dataset.labels = {label: list(set(values)) for label, values in dataset.get_metadata(label_names).items()}

        return dataset

    @classmethod
    def build(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()), ['metadata_file', 'name', 'repertoire_ids', 'metadata_fields'],
                                               RepertoireDataset.__name__, "repertoire dataset")
        repertoires = []
        metadata_df = pd.read_csv(kwargs['metadata_file'], comment=Constants.COMMENT_SIGN)
        for index, row in metadata_df.iterrows():
            filename = Path(kwargs['metadata_file']).parent / row['filename']
            if not filename.is_file() and 'repertoires' in str(filename):
                filename = filename.parent.parent / Path(row['filename']).name
            repertoire = Repertoire(data_filename=filename,
                                    metadata_filename=filename.parent / f'{filename.stem}_metadata.yaml',
                                    identifier=row['identifier'] if 'identifier' in row else uuid.uuid4().hex)
            repertoires.append(repertoire)

        if "repertoire_id" in kwargs.keys() and "repertoires" not in kwargs.keys() and kwargs['repertoire_id'] is not None:
            assert all(rep.identifier == kwargs['repertoire_id'][i] for i, rep in enumerate(repertoires)), \
                f"{RepertoireDataset.__name__}: repertoire ids from the iml_dataset file and metadata file don't match for the dataset " \
                f"{kwargs['name']} with identifier {kwargs['identifier']}."

        return RepertoireDataset(**{**kwargs, **{"repertoires": repertoires}})

    def __init__(self, labels: dict = None, encoded_data: EncodedData = None, repertoires: list = None, identifier: str = None,
                 metadata_file: Path = None, name: str = None, metadata_fields: list = None, repertoire_ids: list = None,
                 example_weights: list = None):
        super().__init__(encoded_data, name, identifier if identifier is not None else uuid.uuid4().hex, labels, example_weights)
        self.metadata_file = Path(metadata_file) if metadata_file is not None else None
        self.metadata_fields = metadata_fields
        self.repertoire_ids = repertoire_ids
        self.repertoires = repertoires

    def clone(self, keep_identifier: bool = False):
        dataset = RepertoireDataset(self.labels, copy.deepcopy(self.encoded_data), copy.deepcopy(self.repertoires),
                                    metadata_file=self.metadata_file, name=self.name)
        if keep_identifier:
            dataset.identifier = self.identifier
        return dataset

    def add_encoded_data(self, encoded_data: EncodedData):
        self.encoded_data = encoded_data

    def get_data(self, batch_size: int = 1):
        return self.repertoires

    def get_batch(self, batch_size: int = 1):
        return self.repertoires

    def get_repertoire(self, index: int = -1, repertoire_identifier: str = "") -> Repertoire:
        assert index != -1 or repertoire_identifier != "", \
            "RepertoireDataset: cannot import_dataset repertoire since the index nor identifier are set."
        return self.repertoires[index] if index != -1 else [rep for rep in self.repertoires if rep.identifier == repertoire_identifier][0]

    def get_example_count(self):
        return len(self.repertoires)

    def get_metadata_fields(self, refresh=False):
        """Returns the list of metadata fields, includes also the fields that will typically not be used as labels,
        like filename or identifier"""
        if self.metadata_fields is None or refresh:
            df = pd.read_csv(self.metadata_file, sep=",", nrows=0, comment=Constants.COMMENT_SIGN)
            self.metadata_fields = df.columns.values.tolist()
        return self.metadata_fields

    def get_label_names(self, refresh=False):
        """Returns the list of metadata fields which can be used as labels; if refresh=True, it reloads the fields
        from disk"""
        all_metadata_fields = set(self.get_metadata_fields(refresh))
        for non_label in ["subject_id", "filename", "repertoire_id", "identifier"]:
            if non_label in all_metadata_fields:
                all_metadata_fields.remove(non_label)

        return all_metadata_fields

    def get_metadata(self, field_names: list, return_df: bool = False):
        """
        A function to get the metadata of the repertoires. It can be useful in encodings or reports when the repertoire information needed is not
        present only in the label chosen for the ML model (e.g., disease), but also other information (e.g., age, HLA).

        Args:
            field_names (list): list of the metadata fields to return; the fields must be present in the metadata files. To find fields available, use :py:obj:`~immuneML.data_model.dataset.RepertoireDataset.RepertoireDataset.get_label_names` function.
            return_df (bool): determines if the results should be returned as a dataframe where each column corresponds to a field or as a dictionary

        Returns:

            a dictionary where keys are fields names and values are lists of field values for each repertoire; alternatively returns the same information in dataframe format

        """
        assert isinstance(self.metadata_file, Path) and self.metadata_file.is_file(), \
            f"RepertoireDataset: for dataset {self.name} (id: {self.identifier}) metadata file is not set properly. The metadata file points to " \
            f"{self.metadata_file}."

        df = pd.read_csv(self.metadata_file, sep=",", usecols=field_names, comment=Constants.COMMENT_SIGN)
        if return_df:
            return df
        else:
            return df.to_dict("list")

    def get_filenames(self):
        """Returns the paths to files in which repertoire information is stored"""
        return [Path(filename) for filename in self.get_metadata(["filename"])["filename"]]

    def _build_new_metadata(self, indices, path: Path) -> Path:
        if self.metadata_file:
            df = pd.read_csv(self.metadata_file, comment=Constants.COMMENT_SIGN)
            df = df.iloc[indices, :]
            df.to_csv(path, index=False)
            return path
        else:
            return None

    def make_subset(self, example_indices, path: Path, dataset_type: str):
        """
        Creates a new dataset object with only those examples (repertoires) available which were given by index in example_indices argument.

        Args:
            example_indices (list): a list of indices of examples (repertoires) to use in the new dataset
            path (Path): a path where to store the newly created dataset
            dataset_type (str): a type of the dataset used as a part of the name of the resulting dataset; the values are defined as constants in :py:obj:`~immuneML.data_model.dataset.Dataset.Dataset`

        Returns:

            a new RepertoireDataset object which includes only the repertoires specified under example_indices

        """

        metadata_file = self._build_new_metadata(example_indices, path / f"{dataset_type}_metadata.csv")
        new_dataset = RepertoireDataset(repertoires=[self.repertoires[i] for i in example_indices], labels=copy.deepcopy(self.labels),
                                        metadata_file=metadata_file, identifier=str(uuid.uuid1()))

        original_example_weights = self.get_example_weights()
        if original_example_weights is not None:
            new_dataset.set_example_weights([original_example_weights[i] for i in example_indices])

        return new_dataset

    def get_repertoire_ids(self) -> list:
        """Returns a list of repertoire identifiers, same as get_example_ids()"""
        if self.repertoire_ids is None:
            self.repertoire_ids = [str(repertoire.identifier) for repertoire in self.repertoires]
        return self.repertoire_ids

    def get_example_ids(self):
        """Returns a list of example identifiers"""
        return self.get_repertoire_ids()

    def get_subject_ids(self):
        """Returns a list of subject identifiers"""
        return self.get_metadata(["subject_id"])["subject_id"]

    def get_data_from_index_range(self, start_index: int, end_index: int):
        return self.repertoires[start_index:end_index+1]



import abc

from immuneML.reports.Report import Report
from immuneML.reports.ReportResult import ReportResult


class ClusteringReport(Report):
    @classmethod
    def build_object(cls, **kwargs):
        pass

    @abc.abstractmethod
    def _generate(self) -> ReportResult:
        pass



from pathlib import Path
from typing import Tuple

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.IO.dataset_import.IReceptorImport import IReceptorImport
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.util.Logger import log, print_log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class ImportParser:

    keyword = "datasets"
    valid_keys = ["format", "params"]

    @staticmethod
    def parse(workflow_specification: dict, symbol_table: SymbolTable, path: Path) -> Tuple[SymbolTable, dict]:

        for key in workflow_specification.keys():
            dataset = ImportParser.parse_dataset(key, workflow_specification[key], path)
            symbol_table.add(key, SymbolType.DATASET, dataset)

        return symbol_table, workflow_specification

    @staticmethod
    @log
    def parse_dataset(key: str, dataset_specs: dict, result_path: Path) -> Dataset:
        location = "ImportParser"

        ParameterValidator.assert_keys(list(dataset_specs.keys()), ImportParser.valid_keys, location, f"datasets:{key}", False)

        valid_formats = ReflectionHandler.all_nonabstract_subclass_basic_names(DataImport, "Import", "IO/dataset_import/")
        ParameterValidator.assert_in_valid_list(dataset_specs["format"], valid_formats, location, "format")

        import_cls = ReflectionHandler.get_class_by_name("{}Import".format(dataset_specs["format"]))
        params = ImportParser._prepare_params(dataset_specs, result_path, key)

        if "is_repertoire" in params:
            ParameterValidator.assert_type_and_value(params["is_repertoire"], bool, location, "is_repertoire")

            if params["is_repertoire"]:
                if import_cls != IReceptorImport:
                    assert "metadata_file" in params, f"{location}: Missing parameter: metadata_file under {key}/params/"
                    ParameterValidator.assert_type_and_value(params["metadata_file"], Path, location, "metadata_file")
            else:
                assert "paired" in params, f"{location}: Missing parameter: paired under {key}/params/"
                ParameterValidator.assert_type_and_value(params["paired"], bool, location, "paired")

                if params["paired"]:
                    assert "receptor_chains" in params, f"{location}: Missing parameter: receptor_chains under {key}/params/"
                    ParameterValidator.assert_in_valid_list(params["receptor_chains"], ["_".join(cp.value) for cp in ChainPair], location, "receptor_chains")

        try:
            dataset = import_cls.import_dataset(params, key)
            dataset.name = key
            ImportParser.log_dataset_info(dataset)
        except KeyError as key_error:
            raise KeyError(f"{key_error}\n\nAn error occurred during parsing of dataset {key}. "
                           f"The keyword {key_error.args[0]} was missing. This either means this argument was "
                           f"not defined under definitions/datasets/{key}/params, or this column was missing from "
                           f"an input data file. ")
        except Exception as ex:
            raise Exception(f"{ex}\n\nAn error occurred while parsing the dataset {key}. See the log above for more details.")

        return dataset

    @staticmethod
    def _prepare_params(dataset_specs: dict, result_path: Path, dataset_name: str):
        params = DefaultParamsLoader.load(ImportParser.keyword, dataset_specs["format"])
        if "params" in dataset_specs.keys():
            params = {**params, **dataset_specs["params"]}
        if "result_path" not in params or params["result_path"] is None:
            params["result_path"] = Path(result_path) / "datasets" / dataset_name
        else:
            params["result_path"] = Path(params["result_path"])

        if "path" in params:
            params["path"] = Path(params["path"])
        if "metadata_file" in params:
            params["metadata_file"] = Path(params["metadata_file"])
        dataset_specs["params"] = params
        return params

    @staticmethod
    def log_dataset_info(dataset: Dataset):
        print_log(f"\nImported {dataset.__class__.__name__.split('Dataset')[0].lower()} dataset {dataset.name}:\n"
                  f"Example count: {dataset.get_example_count()}\n"
                  f"Labels: {dataset.get_label_names()}", True)



import json
from enum import Enum

import numpy as np

class NumpyHelper:

    SIMPLE_TYPES = [str, int, float, bool, np.str_, np.int_, np.float_, np.bool_]

    @staticmethod
    def group_structured_array_by(data, field):
        for col in data.dtype.names:
            data[col][np.argwhere(data[col] == None)] = ""
        sorted_data = np.sort(data, order=[field], axis=0)
        grouped_lists = np.split(sorted_data, np.cumsum(np.unique(sorted_data[field], return_counts=True)[1])[:-1])
        return grouped_lists

    @staticmethod
    def is_simple_type(t):
        """returns if the type t is string or a number so that it does not use pickle if serialized"""
        return t in NumpyHelper.SIMPLE_TYPES

    @staticmethod
    def get_numpy_representation(obj):
        """converts object to representation that can be stored without pickle enables in numpy arrays; if it is an object or a dict,
        it will be serialized to a json string"""

        if obj is None:
            return ''
        elif type(obj) in NumpyHelper.SIMPLE_TYPES:
            return obj
        elif isinstance(obj, Enum):
            return obj.name
        elif not isinstance(obj, dict) and not isinstance(obj, list) and not isinstance(obj, Enum):
            representation = vars(obj)
        else:
            representation = obj

        return json.dumps(representation, default=lambda x: x.name if isinstance(x, Enum) else str(x))

    @staticmethod
    def is_nan_or_empty(value):
        return value == 'nan' or value is None or (not isinstance(value, str) and np.isnan(value)) or value == ''



from dataclasses import dataclass
from pathlib import Path

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.reports.Report import Report


class EncodingReport(Report):
    """
    Encoding reports show some type of features or statistics about an encoded dataset, or may in some cases
    export relevant sequences or tables.

    When running the :ref:`TrainMLModel` instruction, encoding reports can be specified inside the 'selection' or 'assessment' specification under the key 'reports/encoding'.
    Example:

    .. indent with spaces
    .. code-block:: yaml

        my_instruction:
            type: TrainMLModel
            selection:
                reports:
                    encoding:
                        - my_encoding_report
                # other parameters...
            assessment:
                reports:
                    encoding:
                        - my_encoding_report
                # other parameters...
            # other parameters...

    Alternatively, when running the :ref:`ExploratoryAnalysis` instruction, encoding reports can be specified under 'report'. Example:

    .. indent with spaces
    .. code-block:: yaml

        my_instruction:
            type: ExploratoryAnalysis
            analyses:
                my_first_analysis:
                    report: my_encoding_report
                    # other parameters...
            # other parameters...

    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None, number_of_processes: int = 1):
        """
        The arguments defined below are set at runtime by the instruction.
        Concrete classes inheriting EncodingReport may include additional parameters that will be set by the user in the form of input arguments.

        dataset (Dataset): an encoded dataset where encoded_data attribute is set to an instance of EncodedData object
        result_path (Path): path where the results will be stored (plots, tables, etc.)
        name (str): user-defined name of the report that will be shown in the HTML overview later
        number_of_processes (int): how many processes should be created at once to speed up the analysis. For personal machines, 4 or 8 is usually a good choice.
        """
        super().__init__(name, number_of_processes)
        self.dataset = dataset
        self.result_path = result_path

    @staticmethod
    def get_title():
        return "Encoding reports"



# quality: gold

import errno
import os
import shutil
import warnings
from pathlib import Path


class PathBuilder:

    @staticmethod
    def build(path, warn_if_exists=False):
        path = Path(path)
        if warn_if_exists and path.is_dir():
            warnings.warn(f"PathBuilder: directory {path} already exists. Writing in the existing directory...", RuntimeWarning)
        else:
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

        return path

    @staticmethod
    def remove_old_and_build(path):
        path = Path(path)
        if path.is_dir():
            shutil.rmtree(path)

        path.mkdir(parents=True)

        return path



import airr
import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.receptor_sequence.SequenceFrameType import SequenceFrameType
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.util.ImportHelper import ImportHelper
from scripts.specification_util import update_docs_per_mapping


class AIRRImport(DataImport):
    """
    Imports data in AIRR format into a Repertoire-, Sequence- or ReceptorDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets or ReceptorDatasets should be used when predicting values for unpaired (single-chain) and paired
    immune receptors respectively, like antigen specificity.

    The AIRR .tsv format is explained here: https://docs.airr-community.org/en/stable/datarep/format.html
    And the AIRR rearrangement schema can be found here: https://docs.airr-community.org/en/stable/datarep/rearrangements.html

    When importing a ReceptorDataset, the AIRR field cell_id is used to determine the chain pairs.

    Specification arguments:

    - path (str): For RepertoireDatasets, this is the path to a directory with AIRR files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. Only the AIRR files included under the column 'filename' are imported into the RepertoireDataset. For setting SequenceDataset metadata, metadata_file is ignored, see metadata_column_mapping instead.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on the identifier given in the AIRR column named 'cell_id'.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor. Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum. If receptor_chains is not provided, the chain pair is automatically detected (only one chain pair type allowed per repertoire).

    - import_productive (bool): Whether productive sequences (with value 'T' in column productive) should be included in the imported sequences. By default, import_productive is True.

    - import_unknown_productivity (bool): Whether sequences with unknown productivity (missing value in column productive) should be included in the imported sequences. By default, import_unknown_productivity is True.

    - import_with_stop_codon (bool): Whether sequences with stop codons (with value 'T' in column stop_codon) should be included in the imported sequences. This only applies if column stop_codon is present. By default, import_with_stop_codon is False.

    - import_out_of_frame (bool): Whether out of frame sequences (with value 'F' in column vj_in_frame) should be included in the imported sequences. This only applies if column vj_in_frame is present. By default, import_out_of_frame is False.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as AIRR uses the IMGT junction. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from AIRR column names to immuneML's internal data representation. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the AIRR file, or using alternative column names). Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS. For AIRR, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

            junction: sequence
            junction_aa: sequence_aa
            locus: chain

    - column_mapping_synonyms (dict): This is a column mapping that can be used if a column could have alternative names. The formatting is the same as column_mapping. If some columns specified in column_mapping are not found in the file, the columns specified in column_mapping_synonyms are instead attempted to be loaded. For AIRR format, there is no default column_mapping_synonyms.

    - metadata_column_mapping (dict): Specifies metadata for Sequence- and ReceptorDatasets. This should specify a mapping similar to column_mapping where keys are AIRR column names and values are the names that are internally used in immuneML as metadata fields. These metadata fields can be used as prediction labels for Sequence- and ReceptorDatasets. This parameter can also be used to specify sequence-level metadata columns for RepertoireDatasets, which can be used by reports. To set prediction label metadata for RepertoireDatasets, see metadata_file instead. For AIRR format, there is no default metadata_column_mapping.

    - separator (str): Column separator, for AIRR this is by default "\\t".


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_airr_dataset:
            format: AIRR
            params:
                path: path/to/files/
                is_repertoire: True # whether to import a RepertoireDataset
                metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                metadata_column_mapping: # metadata column mapping AIRR: immuneML for Sequence- or ReceptorDatasetDataset
                    airr_column_name1: metadata_label1
                    airr_column_name2: metadata_label2
                import_productive: True # whether to include productive sequences in the dataset
                import_with_stop_codon: False # whether to include sequences with stop codon in the dataset
                import_out_of_frame: False # whether to include out of frame sequences in the dataset
                import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                import_empty_nt_sequences: True # keep sequences even if the `sequences` column is empty (provided that other fields are as specified here)
                import_empty_aa_sequences: False # remove all sequences with empty `sequence_aa` column
                # Optional fields with AIRR-specific defaults, only change when different behavior is required:
                separator: "\\t" # column separator
                region_type: IMGT_CDR3 # what part of the sequence to import
                column_mapping: # column mapping AIRR: immuneML
                    junction: sequence
                    junction_aa: sequence_aa
                    locus: chain

    """

    @staticmethod
    def import_dataset(params: dict, dataset_name: str) -> Dataset:
        return ImportHelper.import_dataset(AIRRImport, params, dataset_name)

    @staticmethod
    def preprocess_dataframe(df: pd.DataFrame, params: DatasetImportParams):
        """
        Function for preprocessing data from a dataframe containing AIRR data, such that:
            - productive sequences, sequences with stop codons or out of frame sequences are filtered according to specification
            - if RegionType is CDR3, the leading C and trailing W are removed from the sequence to match the CDR3 definition
            - if no chain column was specified, the chain is extracted from the v gene name
            - the allele information is removed from the V and J genes
        """
        if "productive" in df.columns:
            df["frame_type"] = SequenceFrameType.UNDEFINED.value
            df.loc[df["productive"]==True, "frame_type"] = SequenceFrameType.IN.value
            df.loc[df["productive"]==False, "frame_type"] = SequenceFrameType.OUT.value
        else:
            df["frame_type"] = None

        if "vj_in_frame" in df.columns:
            df.loc[df["vj_in_frame"]==True, "frame_type"] = SequenceFrameType.IN.value
        if "stop_codon" in df.columns:
            df.loc[df["stop_codon"]==True, "frame_type"] = SequenceFrameType.STOP.value

        if "productive" in df.columns:
            frame_type_list = ImportHelper.prepare_frame_type_list(params)
            df = df[df["frame_type"].isin(frame_type_list)]

        if params.region_type == RegionType.IMGT_CDR3:
            if 'sequence' in df.columns:
                df.drop(['sequence'], axis=1, inplace=True)
            if "cdr3" in df.columns or "cdr3_aa" in df.columns:
                df.rename(columns={"cdr3": "sequence", "cdr3_aa": "sequence_aa"}, inplace=True)
                df.loc[:, 'region_type'] = params.region_type.name
            elif "junction" in df.columns or "junction_aa" in df.columns:
                df.rename(columns={'junction': 'sequence', 'junction_aa': 'sequence_aa'}, inplace=True)
                ImportHelper.junction_to_cdr3(df, params.region_type)
            else:
                df.loc[:, 'region_type'] = params.region_type.name
        else:
            df.loc[:, "region_type"] = params.region_type.name
        # todo else: support "full_sequence" import through regiontype?

        ImportHelper.drop_empty_sequences(df, params.import_empty_aa_sequences, params.import_empty_nt_sequences)
        ImportHelper.drop_illegal_character_sequences(df, params.import_illegal_characters, import_with_stop_codon=params.import_with_stop_codon)
        ImportHelper.load_chains(df)

        return df

    @staticmethod
    def alternative_load_func(filename, params):
        df = airr.load_rearrangement(filename)
        ImportHelper.standardize_none_values(df)
        df.dropna(axis="columns", how="all", inplace=True)
        return df

    @staticmethod
    def import_receptors(df, params):
        df["receptor_id"] = df["cell_id"]
        return ImportHelper.import_receptors(df, params)

    @staticmethod
    def get_documentation():
        doc = str(AIRRImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")
        repertoire_fields = list(Repertoire.FIELDS)
        repertoire_fields.remove("region_type")

        mapping = {
            "Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS": f"Valid immuneML fields that can be specified here are {repertoire_fields}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import logging
from typing import List, Tuple

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from sklearn import metrics
from sklearn.metrics import precision_recall_curve

from immuneML.environment.Constants import Constants
from immuneML.environment.Label import Label
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.multi_dataset_reports.MultiDatasetReport import MultiDatasetReport
from immuneML.util.PathBuilder import PathBuilder


class PerformanceOverview(MultiDatasetReport):
    """
    PerformanceOverview report creates an ROC plot and precision-recall plot for optimal trained models on multiple datasets. The labels on the plots
    are the names of the datasets, so it might be good to have user-friendly names when defining datasets that are still a combination of
    letters, numbers and the underscore sign.

    This report can be used only with MultiDatasetBenchmarkTool as it will plot ROC and PR curve for trained models across datasets. Also, it requires
    the task to be immune repertoire classification and cannot be used for receptor or sequence classification. Furthermore, it uses predictions on
    the test dataset to assess the performance and plot the curves. If the parameter refit_optimal_model is set to True, all data will be used to fit
    the optimal model, so there will not be a test dataset which can be used to assess performance and the report will not be generated.

    If datasets have the same number of examples, the baseline PR curve will be plotted as described in this publication:
    Saito T, Rehmsmeier M. The Precision-Recall Plot Is More Informative than the ROC Plot When Evaluating Binary Classifiers on Imbalanced Datasets.
    PLOS ONE. 2015;10(3):e0118432. doi:10.1371/journal.pone.0118432

    If the datasets have different number of examples, the baseline PR curve will not be plotted.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        reports:
            my_performance_report: PerformanceOverview

    """

    @classmethod
    def build_object(cls, **kwargs):
        return PerformanceOverview(**kwargs)

    def _generate(self) -> ReportResult:

        self.result_path = PathBuilder.build(self.result_path / self.name)

        assert all(self.instruction_states[0].label_configuration.get_labels_by_name() == state.label_configuration.get_labels_by_name() and
                   self.instruction_states[0].label_configuration.get_label_values(
                       self.instruction_states[0].label_configuration.get_labels_by_name()[0]) ==
                   state.label_configuration.get_label_values(state.label_configuration.get_labels_by_name()[0])
                   for state in self.instruction_states), \
            "PerformanceOverview: there is a difference in labels between instructions, the plots cannot be created."
        assert len(self.instruction_states[0].label_configuration.get_labels_by_name()) == 1, \
            'PerformanceOverview: multiple labels were provided, but only one can be used in this report.'

        assert all(state.refit_optimal_model is False for state in self.instruction_states), \
            f"{PerformanceOverview.__name__}: no test datasets were available to assess the performance of optimal models as they were refitted on " \
            f"the full datasets. No reports will be generated."

        label = self.instruction_states[0].label_configuration.get_label_objects()[0]

        optimal_hp_items = [list(state.optimal_hp_items.values())[0] for state in self.instruction_states]

        colors = px.colors.sequential.Viridis[::2][::-1]
        figure_auc, table_aucs = self.plot_roc(optimal_hp_items, label, colors)
        figure_pr, table_pr = self.plot_precision_recall(optimal_hp_items, label, colors)

        return ReportResult(name=self.name,
                            info="A ROC plot and a precision-recall plot for optimal trained models on multiple datasets.",
                            output_figures=[figure_auc, figure_pr], output_tables=table_aucs + table_pr)

    def plot_roc(self, optimal_hp_items, label: Label, colors) -> Tuple[ReportOutput, List[ReportOutput]]:
        report_data_outputs = []
        figure = go.Figure()

        figure.add_trace(go.Scatter(x=[0, 1], y=[0, 1], mode='lines', name='baseline', line=dict(color=Constants.PLOTLY_BLACK, dash='dash'),
                                    hoverinfo="skip"))

        for index, item in enumerate(optimal_hp_items):
            if item.test_predictions_path is None:
                logging.warning(f'{PerformanceOverview.__name__}: there are no test predictions for dataset '
                                f'{self.instruction_states[index].dataset.name}, skipping this dataset when generating performance overview...')
            else:

                df = pd.read_csv(item.test_predictions_path)
                true_class = df[f"{label.name}_true_class"].values
                predicted_class = df[f"{label.name}_{label.positive_class}_proba"].values
                fpr, tpr, _ = metrics.roc_curve(y_true=true_class, y_score=predicted_class)
                auc = metrics.roc_auc_score(true_class, predicted_class)
                name = self.instruction_states[index].dataset.name + f' (AUC = {round(auc, 2)})'
                figure.add_trace(go.Scatter(x=fpr, y=tpr, mode='lines', name=name, marker=dict(color=colors[index], line=dict(width=3)), hoverinfo="skip"))

                data_path = self.result_path / f"roc_curve_data_{name}.csv"
                pd.DataFrame({"FPR": fpr, "TPR": tpr}).to_csv(data_path, index=False)
                report_data_outputs.append(ReportOutput(data_path, f'ROC curve data for dataset {name} (csv)'))

        figure_path = self.result_path / "roc_curve.html"
        figure.update_layout(template='plotly_white', xaxis_title='false positive rate', yaxis_title='true positive rate')
        figure.write_html(str(figure_path))

        return ReportOutput(figure_path, 'ROC curve'), report_data_outputs

    def plot_precision_recall(self, optimal_hp_items: list, label: Label, colors):
        report_data_outputs = []
        figure = go.Figure()

        for index, item in enumerate(optimal_hp_items):
            df = pd.read_csv(item.test_predictions_path)

            true_class = df[f"{label.name}_true_class"].values
            predicted_proba = df[f"{label.name}_{label.positive_class}_proba"].values
            precision, recall, _ = precision_recall_curve(y_true=true_class, probas_pred=predicted_proba)
            name = self.instruction_states[index].dataset.name
            figure.add_trace(go.Scatter(x=recall, y=precision, mode='lines', name=name, marker=dict(color=colors[index], line=dict(width=3)),
                                        hoverinfo="skip"))

            data_path = self.result_path / f"precision_recall_data_{name}.csv"
            pd.DataFrame({"precision": precision, "recall": recall}).to_csv(data_path, index=False)
            report_data_outputs.append(ReportOutput(data_path, f'precision-recall curve data for dataset {name}'))

        figure_path = self.result_path / "precision_recall_curve.html"
        figure.update_layout(template='plotly_white', xaxis_title="recall", yaxis_title="precision")
        figure.write_html(str(figure_path))

        return ReportOutput(figure_path, 'precision-recall curve'), report_data_outputs



from pathlib import Path

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.environment import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.workflows.steps.StepParams import StepParams


class MLMethodTrainerParams(StepParams):

    def __init__(self, method: MLMethod, dataset: Dataset, result_path: Path, label: Label, model_selection_cv: bool,
                 model_selection_n_folds: int, cores_for_training: int, train_predictions_path: Path, ml_details_path: Path,
                 optimization_metric: str):
        self.method = method
        self.result_path = result_path
        self.dataset = dataset
        self.label = label
        self.model_selection_cv = model_selection_cv
        self.model_selection_n_folds = model_selection_n_folds
        self.cores_for_training = cores_for_training
        self.train_predictions_path = train_predictions_path
        self.ml_details_path = ml_details_path
        self.optimization_metric = optimization_metric



import numpy as np
import pandas as pd
import yaml

from immuneML.data_model.encoded_data.EncodedData import EncodedData


def load_encoded_data(labels_path: str, encoding_details_path: str, design_matrix_path: str) -> EncodedData:
    """
    Utility function for adding ML methods; if one encodes data using immuneML through YAML specification and exports the
    encoded data using DesignMatrixExporter, this function can be used to import the data and return it in the format
    it would be available if the ML method was called from within immuneML

    Args:

        labels_path (str): path to labels file as exported by the DesignMatrixExporter
        encoding_details_path (str): path to the details file, where example_ids, feature_names and the encoding name will be imported from
        design_matrix_path (str): path to csv or npy file where the design matrix is stored

    Returns:

        EncodedData object as it would be provided to an ML method within immuneML

    """
    # read the data from these files
    # TODO: support H5py
    examples = pd.read_csv(design_matrix_path).values if design_matrix_path.endswith(".csv") else np.load(design_matrix_path, allow_pickle=True)
    labels = pd.read_csv(labels_path).to_dict('list')

    with open(encoding_details_path, "r") as file:
        encoding_details = yaml.safe_load(file)

    # create an EncodedData object which can be used as an input argument for the fit or predict functions
    encoded_data = EncodedData(examples=examples, labels=labels,
                               example_ids=encoding_details['example_ids'],
                               feature_names=encoding_details['feature_names'],
                               encoding=encoding_details['encoding'])

    return encoded_data



# quality: gold

from pathlib import Path

from gensim.models import Word2Vec

from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.encodings.word2vec.model_creator.ModelCreatorStrategy import ModelCreatorStrategy
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.KmerHelper import KmerHelper


class KmerPairModelCreator(ModelCreatorStrategy):

    def create_model(self, dataset: RepertoireDataset, k: int, vector_size: int, batch_size: int, model_path: Path, sequence_type: SequenceType):

        model = Word2Vec(vector_size=vector_size, min_count=1, window=self.window)  # creates an empty model
        all_kmers = KmerHelper.create_all_kmers(k=k, alphabet=EnvironmentSettings.get_sequence_alphabet())
        all_kmers = [[kmer] for kmer in all_kmers]
        model.build_vocab(all_kmers)

        for kmer in all_kmers:
            sentences = KmerHelper.create_kmers_within_HD(kmer=kmer[0],
                                                          alphabet=EnvironmentSettings.get_sequence_alphabet(),
                                                          distance=1)
            model.train(corpus_iterable=sentences, total_words=len(all_kmers), epochs=model.epochs)

        model.save(str(model_path))

        return model



from dataclasses import dataclass
from typing import Dict, List

import numpy as np
import pandas as pd

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.reports.ReportResult import ReportResult


@dataclass
class ClusteringSetting:
    encoder: DatasetEncoder
    encoder_params: dict
    encoder_name: str
    clustering_method: ClusteringMethod
    clustering_params: dict
    clustering_method_name: str
    dim_reduction_method: DimRedMethod = None
    dim_red_params: dict = None
    dim_red_name: str = None

    def get_key(self) -> str:
        key = self.encoder_name
        if self.dim_red_name:
            key += f"_{self.dim_red_name}"
        key += f"_{self.clustering_method_name}"
        return key

    def __str__(self):
        return self.get_key()


@dataclass
class ClusteringItem:
    dataset: Dataset = None
    method: ClusteringMethod = None
    encoder: DatasetEncoder = None
    performance: pd.DataFrame = None
    predictions: np.ndarray = None
    cl_setting: ClusteringSetting = None




from collections import Counter
from multiprocessing.pool import Pool

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.caching.CacheObjectType import CacheObjectType
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.KmerFrequencyEncoder import KmerFrequencyEncoder
from immuneML.util.Logger import log


class KmerFreqRepertoireEncoder(KmerFrequencyEncoder):

    @log
    def _encode_new_dataset(self, dataset, params: EncoderParams):

        encoded_data = self._encode_data(dataset, params)

        encoded_dataset = RepertoireDataset(repertoires=dataset.repertoires,
                                            encoded_data=encoded_data,
                                            labels=dataset.labels,
                                            metadata_file=dataset.metadata_file)

        return encoded_dataset

    @log
    def _encode_examples(self, dataset, params: EncoderParams):

        arguments = [(repertoire, params) for repertoire in dataset.repertoires]

        with Pool(params.pool_size) as pool:
            repertoires = pool.starmap(self.get_encoded_repertoire, arguments)

        encoded_repertoire_list, repertoire_names, labels, feature_annotation_names = zip(*repertoires)

        encoded_labels = {k: [dic[k] for dic in labels] for k in labels[0]} if params.encode_labels else None

        feature_annotation_names = feature_annotation_names[0]

        return list(encoded_repertoire_list), list(repertoire_names), encoded_labels, feature_annotation_names

    def get_encoded_repertoire(self, repertoire, params: EncoderParams):
        params.model = vars(self)

        return CacheHandler.memo_by_params((("encoding_model", params.model), ("type", "kmer_encoding"),
                                            ("labels", params.label_config.get_labels_by_name()),
                                            ("repertoire_id", repertoire.identifier)),
                                           lambda: self.encode_repertoire(repertoire, params), CacheObjectType.ENCODING_STEP)

    def encode_repertoire(self, repertoire, params: EncoderParams):
        counts = Counter()
        sequence_encoder = self._prepare_sequence_encoder()
        feature_names = sequence_encoder.get_feature_names(params)
        for sequence in repertoire.sequences:
            counts = self._encode_sequence(sequence, params, sequence_encoder, counts)

        label_config = params.label_config
        labels = dict() if params.encode_labels else None

        if labels is not None:
            for label_name in label_config.get_labels_by_name():
                label = repertoire.metadata[label_name]
                labels[label_name] = label

        # TODO: refactor this not to return 4 values but e.g. a dict or split into different functions?
        return counts, repertoire.identifier, labels, feature_names



# quality: gold

import abc
from pathlib import Path

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.receptor.RegionType import RegionType


class DataExporter(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def export(dataset: Dataset, path: Path, number_of_processes: int = 1):
        pass



from random import randrange

import numpy as np
import pandas as pd

from immuneML.analysis.criteria_matches.BooleanType import BooleanType
from immuneML.analysis.criteria_matches.DataType import DataType
from immuneML.analysis.criteria_matches.OperationType import OperationType


class CriteriaMatcher:
    """
    Takes a data frame (for example, repertoire or feature annotations) and criteria and allowed values
    as input and returns a list of boolean values indicating a match or not for each row
    """

    def match(self, criteria, data):
        """
        filter_params = {
            "type": BooleanType.OR,
            "operands": [
                {
                    "type": BooleanType.AND,
                    "operands": [
                        {
                            "type": OperationType.IN,
                            "allowed_values": ["GAD", "PPI"],
                            "value": {
                                "type": DataType.COLUMN,
                                "name": "matching_specificity"
                            }
                        },
                        {
                            "type": OperationType.LESS_THAN,
                            "threshold": 0.001,
                            "value": {
                                "type": DataType.COLUMN,
                                "name": "p_val"
                            }
                        },
                    ]
                },
                {
                    "type": BooleanType.AND,
                    "operands": [
                        {
                            "type": OperationType.IN,
                            "allowed_values": ["yes"],
                            "value": {
                                "type": DataType.COLUMN,
                                "name": "a"
                            }
                        },
                        {
                            "type": OperationType.GREATER_THAN,
                            "threshold": 0.5,
                            "value": {
                                "type": DataType.COLUMN,
                                "name": "odds_ratio"
                            }
                        },
                    ]
                },
            ]
        }
            The order of matching is done from the innermost list to the outermost list. Each column has criteria specified
            with a dictionary with keys "column", "type", and "value". "column" must be present within the data being
            matched to. Allowed "type"s are "in", "less_than", and "greater_than". Between each column criteria, an operator
            must be specified to specify how the criteria are to be combined. To force an order of operations, it can be
            nested within a list. Innermost lists are done first, and recursively simplified until it reaches the outermost
            list to match criteria.
        :param data:
        :return:
        """
        return CriteriaMatcher.parse_criteria(criteria, data)

    @staticmethod
    def evaluate_in(data: pd.Series, criteria: dict):
        result = data.isin(criteria["allowed_values"])
        return result.values

    @staticmethod
    def evaluate_not_in(data: pd.Series, criteria: dict):
        result = ~data.isin(criteria["allowed_values"])
        return result.values

    @staticmethod
    def evaluate_not_na(data: pd.Series, criteria: dict):
        result = data.notna()
        return result.values

    @staticmethod
    def evaluate_greater_than(data: pd.Series, criteria: dict):
        result = data > criteria["threshold"]
        return result.values

    @staticmethod
    def evaluate_less_than(data: pd.Series, criteria: dict):
        result = data < criteria["threshold"]
        return result.values

    @staticmethod
    def evaluate_top_n(data: pd.Series, criteria: dict):
        top_n = data.values.argsort()[(-1 * criteria["number"]):][::-1]
        result = [i in top_n for i in range(data.size)]
        return np.array(result)

    @staticmethod
    def evaluate_random_n(data: pd.Series, criteria: dict):
        random_n = [randrange(0, data.size) for i in range(criteria["number"])]
        result = [i in random_n for i in range(data.size)]
        return np.array(result)

    @staticmethod
    def evaluate_and(booleans: list):
        result = np.logical_and.reduce(booleans)
        return result

    @staticmethod
    def evaluate_or(booleans: list):
        result = np.logical_or.reduce(booleans)
        return result

    @staticmethod
    def evaluate_column(data: pd.Series, name: str) -> pd.Series:
        return data[name]

    @staticmethod
    def parse_criteria(criteria, data):
        if criteria["type"] in DataType:
            return data[criteria["name"]]
        elif criteria["type"] in OperationType:
            operation = getattr(CriteriaMatcher, "evaluate_" + criteria["type"].name.lower())
            return operation(CriteriaMatcher.parse_criteria(criteria["value"], data), criteria)
        elif criteria["type"] in BooleanType:
            operation = getattr(CriteriaMatcher, "evaluate_" + criteria["type"].name.lower())
            booleans = []
            for operand in criteria["operands"]:
                booleans.append(CriteriaMatcher.parse_criteria(operand, data))
            return operation(booleans)



from enum import Enum
from typing import List

from immuneML.data_model.receptor.BCKReceptor import BCKReceptor
from immuneML.data_model.receptor.BCReceptor import BCReceptor
from immuneML.data_model.receptor.TCABReceptor import TCABReceptor
from immuneML.data_model.receptor.TCGDReceptor import TCGDReceptor
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain


class ChainPair(Enum):

    TRA_TRB = (Chain.ALPHA.value, Chain.BETA.value)
    TRG_TRD = (Chain.GAMMA.value, Chain.DELTA.value)
    IGH_IGL = (Chain.HEAVY.value, Chain.LIGHT.value)
    IGH_IGK = (Chain.HEAVY.value, Chain.KAPPA.value)

    @staticmethod
    def get_chain_pair(chains: List[Chain]):
        """Given a list of 2 chain objects, returns the relevant ChainPair"""
        assert len(chains) == 2, f"ChainPair: expected 2 chains, received {len(chains)}"
        assert type(chains[0]) == Chain and type(chains[1]) == Chain, f"ChainPair: expected Chain objects, received {type(chains[0])} and {type(chains[1])}"

        if Chain.ALPHA in chains and Chain.BETA in chains:
            return ChainPair.TRA_TRB
        elif Chain.GAMMA in chains and Chain.DELTA in chains:
            return ChainPair.TRG_TRD
        elif Chain.HEAVY in chains and Chain.LIGHT in chains:
            return ChainPair.IGH_IGL
        elif Chain.HEAVY in chains and Chain.KAPPA in chains:
            return ChainPair.IGH_IGK
        else:
            raise ValueError(f"ChainPair: illegal chain combination: {chains[0]} and {chains[1]}")

    def get_appropriate_receptor_class(self):
        if self.name == 'TRA_TRB':
            return TCABReceptor
        elif self.name == 'TRG_TRD':
            return TCGDReceptor
        elif self.name == 'IGH_IGL':
            return BCReceptor
        elif self.name == 'IGH_IGK':
            return BCKReceptor
        else:
            raise RuntimeError(f"No receptor class defined for {self}.")



import abc

from immuneML.data_model.dataset.Dataset import Dataset


class DimRedMethod:

    def __init__(self, name: str = None):
        self.method = None
        self.name = name

    @abc.abstractmethod
    def fit(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def transform(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def fit_transform(self, dataset: Dataset):
        pass

    @classmethod
    def get_title(cls):
        return "Dimensionality Reduction"



from enum import Enum


class Chain(Enum):
    ALPHA = "TRA"
    BETA = "TRB"
    GAMMA = "TRG"
    DELTA = "TRD"
    HEAVY = "IGH"
    LIGHT = "IGL"
    KAPPA = "IGK"

    @staticmethod
    def get_chain(item: str):
        if type(item) is Chain:
            return item
        elif item.upper() in ["TRA", "TCRA", "A", "ALPHA", "TCRA"]:
            return Chain.ALPHA
        elif item.upper() in ["TRB", "TCRB", "B", "BETA", "TCRB"]:
            return Chain.BETA
        elif item.upper() in ["TRD", "TCRD", "D", "DELTA", "TCRD"]:
            return Chain.DELTA
        elif item.upper() in ["TRG", "TCRG", "G", "GAMMA", "TCRG"]:
            return Chain.GAMMA
        elif item.upper() in ["IGH", "H", "HEAVY"]:
            return Chain.HEAVY
        elif item.upper() in ["IGL", "L", "LIGHT", "LAMBDA"]:
            return Chain.LIGHT
        elif item.upper() in ["IGK", "K", "KAPPA"]:
            return Chain.KAPPA
        else:
            return None

    def __str__(self):
        return self.name

    def to_string(self):
        return self.value



import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.receptor_sequence.SequenceFrameType import SequenceFrameType
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.util.ImportHelper import ImportHelper
from scripts.specification_util import update_docs_per_mapping


class TenxGenomicsImport(DataImport):
    """
    Imports data from the 10x Genomics Cell Ranger analysis pipeline into a Repertoire-, Sequence- or ReceptorDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets or ReceptorDatasets should be used when predicting values for unpaired (single-chain) and paired
    immune receptors respectively, like antigen specificity.

    The files that should be used as input are named 'Clonotype consensus annotations (CSV)',
    as described here: https://support.10xgenomics.com/single-cell-vdj/software/pipelines/latest/output/annotation#consensus

    Note: by default the 10xGenomics field 'umis' is used to define the immuneML field counts. If you want to use the 10x Genomics
    field reads instead, this can be changed in the column_mapping (set reads: counts).
    Furthermore, the 10xGenomics field clonotype_id is used for the immuneML field cell_id.


    Specification arguments:

    - path (str): For RepertoireDatasets, this is the path to a directory with 10xGenomics files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. For setting Sequence- or ReceptorDataset metadata, metadata_file is ignored, see metadata_column_mapping instead.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on the identifier given in the 10xGenomics column named 'clonotype_id'.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor.  Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum. If receptor_chains is not provided, the chain pair is automatically detected (only one chain pair type allowed per repertoire).

    - import_productive (bool): Whether productive sequences (with value 'True' in column productive) should be included in the imported sequences. By default, import_productive is True.

    - import_unproductive (bool): Whether productive sequences (with value 'Fale' in column productive) should be included in the imported sequences. By default, import_unproductive is False.

    - import_unknown_productivity (bool): Whether sequences with unknown productivity (missing or 'NA' value in column productive) should be included in the imported sequences. By default, import_unknown_productivity is True.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as 10xGenomics uses IMGT junction as CDR3. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from 10xGenomics column names to immuneML's internal data representation. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the 10xGenomics file, or using alternative column names). Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS. For 10xGenomics, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

                cdr3: sequence_aa
                cdr3_nt: sequence
                v_gene: v_call
                j_gene: j_call
                umis: duplicate_count
                clonotype_id: cell_id
                consensus_id: sequence_id

    - column_mapping_synonyms (dict): This is a column mapping that can be used if a column could have alternative names. The formatting is the same as column_mapping. If some columns specified in column_mapping are not found in the file, the columns specified in column_mapping_synonyms are instead attempted to be loaded. For 10xGenomics format, there is no default column_mapping_synonyms.

    - metadata_column_mapping (dict): Specifies metadata for Sequence- and ReceptorDatasets. This should specify a mapping similar to column_mapping where keys are 10xGenomics column names and values are the names that are internally used in immuneML as metadata fields. These metadata fields can be used as prediction labels for Sequence- and ReceptorDatasets. This parameter can also be used to specify sequence-level metadata columns for RepertoireDatasets, which can be used by reports. To set prediction label metadata for RepertoireDatasets, see metadata_file instead. For 10xGenomics format, there is no default metadata_column_mapping.

    - separator (str): Column separator, for 10xGenomics this is by default ",".


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_10x_dataset:
            format: 10xGenomics
            params:
                path: path/to/files/
                is_repertoire: True # whether to import a RepertoireDataset
                metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                paired: False # whether to import SequenceDataset (False) or ReceptorDataset (True) when is_repertoire = False
                receptor_chains: TRA_TRB # what chain pair to import for a ReceptorDataset
                metadata_column_mapping: # metadata column mapping 10xGenomics: immuneML for SequenceDataset
                    tenx_column_name1: metadata_label1
                    tenx_column_name2: metadata_label2
                import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                import_empty_aa_sequences: False # filter out sequences if they don't have sequence_aa set
                # Optional fields with 10xGenomics-specific defaults, only change when different behavior is required:
                separator: "," # column separator
                region_type: IMGT_CDR3 # what part of the sequence to import
                column_mapping: # column mapping 10xGenomics: immuneML
                    cdr3: sequence_aa
                    cdr3_nt: sequence
                    v_gene: v_call
                    j_gene: j_call
                    umis: duplicate_count
                    clonotype_id: cell_id
                    consensus_id: sequence_id

    """

    @staticmethod
    def import_dataset(params: dict, dataset_name: str) -> Dataset:
        return ImportHelper.import_dataset(TenxGenomicsImport, params, dataset_name)

    @staticmethod
    def preprocess_dataframe(df: pd.DataFrame, params: DatasetImportParams):
        df["frame_type"] = SequenceFrameType.UNDEFINED.value
        df.loc[df['productive']=="True", "frame_type"] = SequenceFrameType.IN.value
        df.loc[df['productive']=="False", "frame_type"] = SequenceFrameType.OUT.value

        allowed_productive_values = []
        if params.import_productive:
            allowed_productive_values.append('True')
        if params.import_unproductive:
            allowed_productive_values.append('False')
        if params.import_unknown_productivity:
            allowed_productive_values.append('')
            allowed_productive_values.append('NA')

        df = df[df.productive.isin(allowed_productive_values)]
        df.drop(columns=["productive"], inplace=True)

        ImportHelper.junction_to_cdr3(df, params.region_type)
        df.loc[:, "region_type"] = params.region_type.name
        ImportHelper.drop_empty_sequences(df, params.import_empty_aa_sequences, params.import_empty_nt_sequences)
        ImportHelper.drop_illegal_character_sequences(df, params.import_illegal_characters, params.import_with_stop_codon)
        ImportHelper.load_chains(df)

        return df

    @staticmethod
    def import_receptors(df, params):
        df["receptor_id"] = df["cell_id"]
        return ImportHelper.import_receptors(df, params)

    @staticmethod
    def get_documentation():
        doc = str(TenxGenomicsImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")
        repertoire_fields = list(Repertoire.FIELDS)
        repertoire_fields.remove("region_type")

        mapping = {
            "Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS": f"Valid immuneML fields that can be specified here are {repertoire_fields}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc





from dataclasses import dataclass
from pathlib import Path

from immuneML.environment.LabelConfiguration import LabelConfiguration


@dataclass
class EncoderParams:
    result_path: Path
    label_config: LabelConfiguration
    pool_size: int = 4
    model: dict = None
    learn_model: bool = True
    encode_labels: bool = True



import warnings

from pathlib import Path
from typing import List

import pandas as pd
import numpy as np


from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.core.HPUtil import HPUtil
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.ml_metrics.MetricUtil import MetricUtil
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.ml_model_application.MLApplicationState import MLApplicationState
from scripts.specification_util import update_docs_per_mapping


class MLApplicationInstruction(Instruction):
    """
    Instruction which enables using trained ML models and encoders on new datasets which do not necessarily have labeled data.
    When the same label is provided as the ML setting was trained for, performance metrics can be computed.

    The predictions are stored in the predictions.csv in the result path in the following format:

    .. list-table::
        :widths: 25 25 25 25
        :header-rows: 1

        * - example_id
          - cmv_predicted_class
          - cmv_1_proba
          - cmv_0_proba
        * - e1
          - 1
          - 0.8
          - 0.2
        * - e2
          - 0
          - 0.2
          - 0.8
        * - e3
          - 1
          - 0.78
          - 0.22


    If the same label that the ML setting was trained for is present in the provided dataset, the 'true' label value
    will be added to the predictions table in addition:

    .. list-table::
        :widths: 20 20 20 20 20
        :header-rows: 1

        * - example_id
          - cmv_predicted_class
          - cmv_1_proba
          - cmv_0_proba
          - cmv_true_class
        * - e1
          - 1
          - 0.8
          - 0.2
          - 1
        * - e2
          - 0
          - 0.2
          - 0.8
          - 0
        * - e3
          - 1
          - 0.78
          - 0.22
          - 0

    Specification arguments:

    - dataset: dataset for which examples need to be classified

    - config_path: path to the zip file exported from MLModelTraining instruction (which includes train ML model, encoder, preprocessing etc.)

    - number_of_processes (int): how many processes should be created at once to speed up the analysis. For personal machines, 4 or 8 is usually a good choice.

    - metrics (list): a list of metrics to compute between the true and predicted classes. These metrics will only be computed when the same label with the same classes is provided for the dataset as the original label the ML setting was trained for.


    Specification example for the MLApplication instruction:

    .. highlight:: yaml
    .. code-block:: yaml

        instruction_name:
            type: MLApplication
            dataset: d1
            config_path: ./config.zip
            metrics:
            - accuracy
            - precision
            - recall
            number_of_processes: 4

    """

    def __init__(self, dataset: Dataset, label_configuration: LabelConfiguration, hp_setting: HPSetting, metrics: List[ClassificationMetric], number_of_processes: int, name: str):

        self.state = MLApplicationState(dataset=dataset, hp_setting=hp_setting, label_config=label_configuration, metrics=metrics, pool_size=number_of_processes, name=name)

    def run(self, result_path: Path):
        self.state.path = PathBuilder.build(result_path / self.state.name)

        dataset = self.state.dataset

        if self.state.hp_setting.preproc_sequence is not None:
            dataset = HPUtil.preprocess_dataset(dataset, self.state.hp_setting.preproc_sequence, self.state.path)

        dataset = HPUtil.encode_dataset(dataset, self.state.hp_setting, self.state.path, learn_model=False, number_of_processes=self.state.pool_size,
                                        label_configuration=self.state.label_config, context={}, encode_labels=False)

        self._write_outputs(dataset)

        return self.state

    def _write_outputs(self, dataset):
        label = self.state.label_config.get_label_objects()[0]

        predictions_df = self._make_predictions_df(dataset, label)
        self.state.predictions_path = self.state.path / "predictions.csv"
        predictions_df.to_csv(self.state.predictions_path, index=False)

        metrics_df = self._apply_metrics_with_warnings(dataset, label, predictions_df)
        if metrics_df is not None:
            self.state.metrics_path = self.state.path / "metrics.csv"
            metrics_df.to_csv(self.state.metrics_path, index=False)


    def _make_predictions_df(self, dataset, label):
        method = self.state.hp_setting.ml_method
        predictions = method.predict(dataset.encoded_data, label)
        predictions_df = pd.DataFrame({"example_id": dataset.get_example_ids()})

        predictions_df[f"{label.name}_predicted_class"] = predictions[label.name]
        predictions_df[f"{label.name}_predicted_class"] = predictions_df[f"{label.name}_predicted_class"].astype(str)

        if type(dataset) == RepertoireDataset:
            predictions_df.insert(0, 'repertoire_file', [repertoire.data_filename.name for repertoire in dataset.get_data()])

        if method.can_predict_proba():
            predictions_proba = method.predict_proba(dataset.encoded_data, label)[label.name]

            for cls in method.get_classes():
                predictions_df[f'{label.name}_{cls}_proba'] = predictions_proba[cls]

        if label.name in dataset.get_label_names():
            predictions_df[f"{label.name}_true_class"] = dataset.get_metadata([label.name])[label.name]
            predictions_df[f"{label.name}_true_class"] = predictions_df[f"{label.name}_true_class"].astype(str)

        return predictions_df

    def _apply_metrics_with_warnings(self, dataset, label, predictions_df):
        if len(self.state.metrics) > 0:
            if label.name in dataset.get_label_names():
                if all([dataset_class in self.state.hp_setting.ml_method.get_classes() for dataset_class in dataset.labels[label.name]]):
                    return self._apply_metrics(label, predictions_df)
                else:
                    warnings.warn(f"MLApplicationInstruction: tried to apply metrics for label {label.name}. "
                                  f"Found class values {dataset.labels[label.name]} in the provided dataset, "
                                  f"but expected classes {self.state.hp_setting.ml_method.get_classes()}.")
            else:
                warnings.warn(f"MLApplicationInstruction: tried to apply metrics for label {label.name}, "
                              f"but the provided dataset only contains information for the following "
                              f"labels: {dataset.get_label_names()}.")

    def _apply_metrics(self, label, predictions_df):
        result = {}
        for metric in self.state.metrics:
            if all([f'{label.name}_{cls}_proba' in predictions_df.columns for cls in label.values]):
                predicted_proba_y = np.vstack([np.array(predictions_df[f'{label.name}_{cls}_proba']) for cls in label.values]).T
            else:
                predicted_proba_y = None

            result[metric.name.lower()] = [MetricUtil.score_for_metric(metric=metric,
                                                                        predicted_y=np.array(predictions_df[f"{label.name}_predicted_class"]),
                                                                        predicted_proba_y=predicted_proba_y,
                                                                        true_y=np.array(predictions_df[f"{label.name}_true_class"]),
                                                                        classes=[str(val) for val in label.values])]
        return pd.DataFrame(result)

    @staticmethod
    def get_documentation():
        doc = str(MLApplicationInstruction.__doc__)
        valid_metrics = str([metric.name.lower() for metric in ClassificationMetric])[1:-1].replace("'", "`")

        mapping = {
            "a list of metrics": f"a list of metrics ({valid_metrics})",
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



from pathlib import Path

import plotly.express as px

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.reports.encoding_reports.FeatureReport import FeatureReport


class FeatureDistribution(FeatureReport):
    """
    Plots a boxplot for each feature in the encoded data matrix. Can be used in combination
    with any encoding and dataset type.
    Each boxplot represents a feature and shows the distribution of values for that feature.
    For example, when :ref:`KmerFrequency` encoder is used, the features are the
    k-mers (AAA, AAC, etc..) and the feature values are the frequencies per k-mer.

    Two modes can be used: in the 'normal' mode there are normal boxplots corresponding to each column of the
    encoded dataset matrix; in the 'sparse' mode all zero cells are eliminated before
    passing the data to the boxplots. If mode is set to 'auto', then it will automatically
    set to 'sparse' if the density of the matrix is below 0.01

    Optional metadata labels can be specified to divide the boxplots into groups based on color, row facets or column facets.
    These labels are specified in the metadata file for repertoire datasets, or as metadata columns for sequence and receptor datasets.

    Alternatively, when only the mean feature values are of interest (as opposed to showing the complete distribution, as done here),
    please consider using :ref:`FeatureValueBarplot` instead.
    When comparing the feature values between two subsets of the data, please use :ref:`FeatureComparison`.


    Specification arguments:

    - color_grouping_label (str): The label that is used to color each bar, at each level of the grouping_label.

    - row_grouping_label (str): The label that is used to group bars into different row facets.

    - column_grouping_label (str): The label that is used to group bars into different column facets.

    - mode (str): either 'normal', 'sparse' or 'auto' (default)

    - x_title (str): x-axis label

    - y_title (str): y-axis label


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_fdistr_report:
            FeatureDistribution:
                mode: sparse

    """

    @classmethod
    def build_object(cls, **kwargs):
        return FeatureDistribution(**kwargs)

    def __init__(self, dataset: Dataset = None, result_path: Path = None, color_grouping_label: str = None,
                 row_grouping_label=None, column_grouping_label=None,
                 mode: str = 'auto', x_title: str = None, y_title: str = None, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, color_grouping_label=color_grouping_label,
                         row_grouping_label=row_grouping_label, column_grouping_label=column_grouping_label,
                         number_of_processes=number_of_processes, name=name)
        self.x_title = x_title if x_title is not None else self.x
        self.y_title = y_title if y_title is not None else "value"
        self.mode = mode
        self.result_name = "feature_distributions"

    def _generate(self):
        result = self._generate_report_result()
        result.info = "Each boxplot represents one feature of the encoded data matrix, and shows the distribution of values for that feature."
        return result

    def _plot(self, data_long_format, mode='sparse') -> ReportOutput:
        sparse_threshold = 0.01

        if self.mode == 'auto':
            if (data_long_format.value == 0).mean() < sparse_threshold:
                self.mode = 'normal'
            else:
                self.mode = 'sparse'

        if self.mode == 'sparse':
            return self._plot_sparse(data_long_format)
        elif self.mode == 'normal':
            return self._plot_normal(data_long_format)

    def _plot_sparse(self, data_long_format) -> ReportOutput:
        columns_to_filter = [self.x, "value"]
        for optional_column in [self.color, self.facet_row, self.facet_column]:
            if optional_column is not None:
                columns_to_filter.append(optional_column)

        data_long_format_filtered = data_long_format.loc[data_long_format.value != 0, columns_to_filter]
        columns_to_filter.remove("value")
        total_counts = data_long_format_filtered.groupby(columns_to_filter, as_index=False).agg(
            {"value": 'sum'})
        data_long_format_filtered = data_long_format_filtered.merge(total_counts,
                                                                    on=self.x,
                                                                    how="left",
                                                                    suffixes=('', '_sum')) \
            .fillna(0) \
            .sort_values(by=self.x) \
            .reset_index(drop=True)

        figure = px.box(data_long_format_filtered, x=self.x, y="value", color=self.color,
                        facet_row=self.facet_row, facet_col=self.facet_column,
                        labels={
                            "value": self.y_title,
                            self.x: self.x_title,
                        }, template='plotly_white',
                        color_discrete_sequence=px.colors.diverging.Tealrose)

        file_path = self.result_path / f"{self.result_name}.html"

        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="feature boxplots")

    def _plot_normal(self, data_long_format) -> ReportOutput:

        figure = px.box(data_long_format, x=self.x, y="value", color=self.color,
                        facet_row=self.facet_row, facet_col=self.facet_column,
                        labels={
                            "value": self.y_title,
                            self.x: self.x_title,
                        }, template='plotly_white',
                        color_discrete_sequence=px.colors.diverging.Tealrose)

        file_path = self.result_path / f"{self.result_name}.html"

        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Distributions of feature values")


##
class DimensionalitReductionReport(EncodingReport):
    def _generate(self):
        """
        Get dim reduced data from self.dataset.encoded...
        make a plot
        return ReportResult
        """
        pass


from pathlib import Path

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.util.Logger import print_log
from immuneML.workflows.instructions.GenModelInstruction import GenModelState, GenModelInstruction


class TrainGenModelState(GenModelState):
    pass


class TrainGenModelInstruction(GenModelInstruction):
    """
    TrainGenModel instruction implements training generative AIRR models on receptor level. Models that can be trained
    for sequence generation are listed under Generative Models section.

    This instruction takes a dataset as input which will be used to train a model, the model itself, and the number of
    sequences to generate to illustrate the applicability of the model. It can also produce reports of the fitted model
    and reports of original and generated sequences.

    To use the generative model previously trained with immuneML, see ApplyGenModel instruction.

    .. note::

        This is an experimental feature in version 3.0.0a1.

    Specification arguments:

    - dataset: dataset to use for fitting the generative model; it has to be defined under definitions/datasets

    - method: which model to fit (defined previously under definitions/ml_methods)

    - number_of_processes (int): how many processes to use for fitting the model

    - gen_examples_count (int): how many examples (sequences, repertoires) to generate from the fitted model

    - reports (list): list of report ids (defined under definitions/reports) to apply after fitting a generative model
      and generating gen_examples_count examples; these can be data reports (to be run on generated examples), ML
      reports (to be run on the fitted model)

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_train_gen_model_inst: # user-defined instruction name
            type: TrainGenModel
            dataset: d1 # defined previously under definitions/datasets
            model: model1 # defined previously under definitions/ml_methods
            gen_examples_count: 100
            number_of_processes: 4
            reports: [data_rep1, ml_rep2]

    """

    MAX_ELEMENT_COUNT_TO_SHOW = 10

    def __init__(self, dataset: Dataset = None, method: GenerativeModel = None, number_of_processes: int = 1,
                 gen_examples_count: int = 100, result_path: Path = None, name: str = None, reports: list = None):
        super().__init__(TrainGenModelState(result_path, name, gen_examples_count), method, reports)
        self.dataset = dataset
        self.number_of_processes = number_of_processes

    def run(self, result_path: Path) -> GenModelState:
        self._set_path(result_path)
        self._fit_model()
        self._save_model()
        self._gen_data()
        self._evaluate_model()
        self._run_reports()

        return self.state

    def _fit_model(self):
        print_log(f"{self.state.name}: starting to fit the model", True)
        self.method.fit(self.dataset, self.state.result_path)
        print_log(f"{self.state.name}: fitted the model", True)

    def _save_model(self):
        self.state.model_path = self.method.save_model(self.state.result_path / 'trained_model/')

    def _evaluate_model(self):
        print("Evaluation is not implemented yet!")



import logging
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import plotly.express as px

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class RepertoireClonotypeSummary(DataReport):
    """
    Shows the number of distinct clonotypes per repertoire in a given dataset as a bar plot.

    Specification arguments:

    - color_by_label (str): name of the label to use to color the plot, e.g., could be disease label, or None

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_clonotype_summary_rep:
          RepertoireClonotypeSummary:
            color_by_label: celiac

    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None, number_of_processes: int = 1, color_by_label: str = None):
        super().__init__(dataset, result_path, name, number_of_processes)
        self.color_by_label = color_by_label

    @classmethod
    def build_object(cls, **kwargs):
        if "color_by_label" in kwargs and kwargs['color_by_label'] is not None:
            ParameterValidator.assert_type_and_value(kwargs['color_by_label'], str, RepertoireClonotypeSummary.__name__, 'color_by_label')

        return RepertoireClonotypeSummary(**kwargs)

    def _generate(self) -> ReportResult:
        assert isinstance(self.dataset, RepertoireDataset), \
            f"{RepertoireClonotypeSummary.__name__}: expected repertoire dataset, but got {type(self.dataset)}."

        PathBuilder.build(self.result_path)
        return self._safe_plot()

    def _plot(self) -> ReportResult:
        clonotypes = pd.DataFrame.from_records(sorted([self._get_clonotype_count_with_label(repertoire) for repertoire in self.dataset.get_data()],
                                                      key=lambda x: x[0]), columns=['clonotype_count', self.color_by_label])
        clonotypes['repertoire'] = list(range(1, self.dataset.get_example_count()+1))
        fig = px.bar(clonotypes, x='repertoire', y='clonotype_count', color=self.color_by_label, title='Clonotype count per repertoire',
                     color_discrete_sequence=px.colors.qualitative.Pastel2)
        fig.update_layout(template="plotly_white", yaxis_title='clonotype count', xaxis_title='repertoires sorted by clonotype count')
        clonotypes.to_csv(self.result_path / 'clonotype_count_per_repertoire.csv')
        fig.write_html(str(self.result_path / 'clonotype_count_per_repertoire.html'))

        return ReportResult(name=self.name, info="Clonotype count per repertoire",
                            output_figures=[ReportOutput(self.result_path / 'clonotype_count_per_repertoire.html')],
                            output_tables=[ReportOutput(self.result_path / 'clonotype_count_per_repertoire.csv')])

    def _get_clonotype_count_with_label(self, repertoire: Repertoire) -> Tuple[int, str]:

        sequences = repertoire.get_attribute('sequences')
        if sequences is None:
            sequences = repertoire.get_sequence_aas()

        sequence_count = sequences.shape[0]
        unique_sequence_count = np.unique(sequences.tolist()).shape[0]
        if sequence_count != unique_sequence_count:
            logging.warning(f"{RepertoireClonotypeSummary.__name__}: for repertoire {repertoire.identifier}, there are {sequence_count} sequences, "
                            f"but {unique_sequence_count} unique sequences.")

        return unique_sequence_count, repertoire.metadata[self.color_by_label] if self.color_by_label else None



from dataclasses import dataclass
from pathlib import Path

@dataclass
class ReportOutput:
    def __init__(self, path: Path, name: str = None):
        self.path = Path(path)
        self.name = name



from pathlib import Path

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.StringHelper import StringHelper
from immuneML.workflows.instructions.subsampling.SubsamplingState import SubsamplingState


class SubsamplingHTMLBuilder:
    """
    A class that will make a HTML file(s) out of SubsamplingState object to show what analysis took place in
    the SubsamplingInstruction.
    """

    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: SubsamplingState) -> str:
        """
        Function that builds the HTML files based on the Subsampling state.
        Arguments:
            state: SubsamplingState object including all details of the Subsampling instruction
        Returns:
             path to the main HTML file (which is located under state.result_path)
        """
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = SubsamplingHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"Subsampling_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "Subsampling.html",
                             template_map=html_map, result_path=result_file)

        return result_file

    @staticmethod
    def make_html_map(state: SubsamplingState, base_path: Path) -> dict:
        html_map = {
            "css_style": Util.get_css_content(SubsamplingHTMLBuilder.CSS_PATH),
            "name": state.name,
            'immuneML_version': MLUtil.get_immuneML_version(),
            "full_specs": Util.get_full_specs_path(base_path),
            "dataset_name": state.dataset.name if state.dataset.name is not None else state.dataset.identifier,
            "labels": [{"label_name": label} for label in state.dataset.get_label_names()],
            "dataset_type": StringHelper.camel_case_to_word_string(type(state.dataset).__name__),
            "example_count": state.dataset.get_example_count(),
            "subsampled_datasets": [{"sub_dataset_iter": i,
                                     "sub_dataset_name": dataset.name,
                                     "dataset_size": f"{dataset.get_example_count()} {type(dataset).__name__.replace('Dataset', 's').lower()}",
                                     "formats": [{"dataset_download_link": item, "format_name": key}
                                                 for key, item in state.subsampled_dataset_paths[dataset.name].items()]}
                                    for i, dataset in enumerate(state.subsampled_datasets, 1)]
        }

        return html_map



from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.analysis.criteria_matches.CriteriaMatcher import CriteriaMatcher
from immuneML.analysis.criteria_matches.CriteriaTypeInstantiator import CriteriaTypeInstantiator
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.preprocessing.filters.Filter import Filter


class MetadataRepertoireFilter(Filter):
    """
    Removes repertoires from a RepertoireDataset based on information stored in the metadata_file.
    Note that this filters out repertoires, not individual sequences, and can thus only be applied to RepertoireDatasets.

    Since this filter changes the number of repertoires (examples for the machine learning task), it cannot be used with
    :ref:`TrainMLModel` instruction. To filter out repertoires, use preprocessing from the :ref:`DatasetExport` instruction that will create
    a new dataset ready to be used for training machine learning models.

    Specification arguments:

    - criteria (dict): a nested dictionary that specifies the criteria for keeping certain columns. See :py:obj:`~immuneML.analysis.criteria_matches.CriteriaMatcher.CriteriaMatcher` for a more detailed explanation.


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    # Example filter that keeps repertoires with values greater than 1 in the "my_column_name" column of the metadata_file
                    MetadataRepertoireFilter:
                        type: GREATER_THAN
                        value:
                            type: COLUMN
                            name: my_column_name
                        threshold: 1

    """

    def __init__(self, criteria: dict, result_path: Path = None):
        super().__init__(result_path)
        self.criteria = CriteriaTypeInstantiator.instantiate(criteria)

    def keeps_example_count(self) -> bool:
        return False

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1):
        self.check_dataset_type(dataset, [RepertoireDataset], "MetadataRepertoireFilter")
        self.result_path = result_path if result_path is not None else self.result_path

        processed_dataset = dataset.clone()
        original_repertoires = processed_dataset.get_data()
        indices = self._get_matching_indices(processed_dataset)
        processed_dataset.repertoires = [original_repertoires[i] for i in indices]
        processed_dataset.metadata_file = self._build_new_metadata(dataset, indices)

        self.check_dataset_not_empty(processed_dataset, "MetadataRepertoireFilter")

        return processed_dataset

    def _get_matching_indices(self, dataset: RepertoireDataset):
        metadata = pd.DataFrame(dataset.get_metadata(None))
        matches = CriteriaMatcher().match(self.criteria, metadata)
        indices = np.where(matches)[0]
        return indices





from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.reports.ReportResult import ReportResult


@dataclass
class HPItem:

    method: MLMethod = None
    encoder: DatasetEncoder = None
    performance: dict = None
    hp_setting: HPSetting = None
    train_predictions_path: Path = None
    test_predictions_path: Path = None
    ml_details_path: Path = None
    ml_settings_export_path: Path = None
    train_dataset: Dataset = None
    test_dataset: Dataset = None
    split_index: int = None
    model_report_results: List[ReportResult] = field(default_factory=list)
    encoding_train_results: List[ReportResult] = field(default_factory=list)
    encoding_test_results: List[ReportResult] = field(default_factory=list)



import abc
from pathlib import Path

from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.environment.SequenceType import SequenceType


class ModelCreatorStrategy(metaclass=abc.ABCMeta):
    """
    Defines how word2vec model can be created by defining different contexts for k-mers
    """

    def __init__(self, epochs: int, window: int):
        self.epochs = epochs
        self.window = window

    @abc.abstractmethod
    def create_model(self, dataset: RepertoireDataset, k: int, vector_size: int, batch_size: int, model_path: Path, sequence_type: SequenceType):
        pass



import copy

from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.simulation.implants.LigoPWM import LigoPWM
from immuneML.simulation.implants.Motif import Motif
from immuneML.simulation.implants.SeedMotif import SeedMotif
from immuneML.util.Logger import log
from immuneML.util.ParameterValidator import ParameterValidator


class MotifParser:

    keyword = "motifs"

    @staticmethod
    def parse(motifs: dict, symbol_table: SymbolTable):

        for key, motif_dict in motifs.items():

            motif_keys = list(motif_dict.keys())
            if "seed" in motif_keys:
                ParameterValidator.assert_keys(motif_keys, ['seed', 'min_gap', 'max_gap', 'hamming_distance_probabilities', 'position_weights', 'alphabet_weights'], "MotifParser", key, exclusive=False)
                motif = MotifParser._parse_seed_motif(key, motif_dict)
            else:
                ParameterValidator.assert_keys(motif_keys, ['file_path', 'threshold'], "MotifParser", key)
                motif = LigoPWM.build(identifier=key, file_path=motif_dict['file_path'], threshold=motif_dict['threshold'])

            symbol_table.add(key, SymbolType.MOTIF, motif)

        return symbol_table, motifs

    @staticmethod
    @log
    def _parse_seed_motif(key: str, motif_item: dict) -> Motif:

        motif_dict = copy.deepcopy(motif_item)
        motif_dict["identifier"] = key

        if 'hamming_distance_probabilities' in motif_dict and motif_dict['hamming_distance_probabilities'] is not None:
            motif_dict['hamming_distance_probabilities'] = {key: float(value) for key, value in motif_dict['hamming_distance_probabilities'].items()}
            assert all(isinstance(key, int) for key in motif_dict['hamming_distance_probabilities'].keys()) \
                   and all(isinstance(val, float) for val in motif_dict['hamming_distance_probabilities'].values()) \
                   and 0.99 <= sum(motif_dict['hamming_distance_probabilities'].values()) <= 1, \
                   "For each possible Hamming distance a probability between 0 and 1 has to be assigned " \
                   "so that the probabilities for all distance possibilities sum to 1."

        motif = SeedMotif(**motif_dict)

        return motif



import copy
import logging
import math
from pathlib import Path
from typing import List

import bionumpy as bnp
import pandas as pd

from immuneML.data_model.SequenceSet import SequenceSet
from immuneML.data_model.bnp_util import bnp_write_to_file, make_dynamic_seq_set_from_objs, prepare_values_for_bnp, \
    write_yaml
from immuneML.data_model.dataset.ElementDataset import ElementDataset
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.util.PathBuilder import PathBuilder


class SequenceDataset(ElementDataset):
    """A dataset class for sequence datasets (single chain). All the functionality is implemented in ElementDataset class, except creating a new
    dataset and obtaining metadata."""

    DEFAULT_FILE_SIZE = 100000

    @classmethod
    def build_from_objects(cls, sequences: List[ReceptorSequence], file_size: int, path: Path, name: str = None,
                           labels: dict = None):

        file_count = math.ceil(len(sequences) / file_size)
        PathBuilder.build(path)

        file_names = [
            path / f"batch{''.join(['0' for i in range(1, len(str(file_count)) - len(str(index)) + 1)])}{index}.tsv"
            for index in range(1, file_count + 1)]

        seq_set_dc, types = make_dynamic_seq_set_from_objs(sequences)

        for index in range(file_count):
            vals = {field_name: [seq.get_attribute(field_name) for seq in sequences[index * file_size:(index + 1) * file_size]]
                    for field_name in types.keys()}
            vals = prepare_values_for_bnp(vals, types)
            sequence_matrix = seq_set_dc(**vals)
            bnp_write_to_file(file_names[index], sequence_matrix)

        metadata = {
            'type_dict': {key: SequenceSet.TYPE_TO_STR[val] for key, val in types.items()},
            'dataset_class': 'SequenceDataset', 'element_class_name': ReceptorSequence.__name__,
            'filenames': [str(file) for file in file_names]
        }
        dataset_file = path / f'dataset_{name}.yaml'
        write_yaml(dataset_file, metadata)

        return SequenceDataset(filenames=file_names, file_size=file_size, name=name, labels=labels,
                               element_class_name=ReceptorSequence.__name__, dataset_file=dataset_file,
                               buffer_type=bnp.io.delimited_buffers.get_bufferclass_for_datatype(seq_set_dc,
                                                                                                 delimiter='\t',
                                                                                                 has_header=True)
                               )

    def __init__(self, **kwargs):
        super().__init__(**{**kwargs, **{'element_class_name': ReceptorSequence.__name__}})

    def get_metadata(self, field_names: list, return_df: bool = False):
        """Returns a dict or an equivalent pandas DataFrame with metadata information under 'custom_params' attribute in SequenceMetadata object
        for every sequence for provided field names"""
        result = {field: [] for field in field_names}
        for sequence in self.get_data():
            for field in field_names:
                result[field].append(sequence.metadata.get_attribute(field))

        for field in field_names:
            if all(item is None for item in result[field]):
                logging.warning(
                    f"{SequenceDataset.__name__}: none of the sequences in the dataset {self.name} have metadata field '{field}'. "
                    f"Returning 'None' instead...")
                result[field] = None

        return pd.DataFrame(result) if return_df else result

    def clone(self, keep_identifier: bool = False):
        dataset = SequenceDataset(labels=self.labels, encoded_data=copy.deepcopy(self.encoded_data),
                                  filenames=copy.deepcopy(self.filenames), dataset_file=self.dataset_file,
                                  file_size=self.file_size, name=self.name)
        if keep_identifier:
            dataset.identifier = self.identifier
        dataset.element_ids = self.element_ids
        return dataset



from pathlib import Path

import logging
import plotly.express as px
import numpy as np
import pandas as pd

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.encodings.motif_encoding.PositionalMotifHelper import PositionalMotifHelper

from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.PathBuilder import PathBuilder


class GroundTruthMotifOverlap(EncodingReport):
    """
    Creates report displaying overlap between learned motifs and groundtruth motifs

    # todo: arguments, yaml spec, explanation of format of highlight motifs file
    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None,
                 number_of_processes: int = 1, groundtruth_motifs_path: str = None):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)
        self.groundtruth_motifs_path = groundtruth_motifs_path

    @classmethod
    def build_object(cls, **kwargs):
        location = GroundTruthMotifOverlap.__name__

        if "groundtruth_motifs_path" in kwargs and kwargs["groundtruth_motifs_path"] is not None:
            PositionalMotifHelper.check_motif_filepath(kwargs["groundtruth_motifs_path"], location, "groundtruth_motifs_path", expected_header="indices\tamino_acids\tn_sequences\n")

        return GroundTruthMotifOverlap(**kwargs)

    def _generate(self):
        PathBuilder.build(self.result_path)

        groundtruth_motifs, implant_rate_dict = self._read_groundtruth_motifs(self.groundtruth_motifs_path)

        learned_motifs = self.dataset.encoded_data.feature_names

        overlap_df = self._generate_overlap(learned_motifs, groundtruth_motifs, implant_rate_dict)
        output_table = self._write_output_table(overlap_df, self.result_path / "ground_truth_motif_overlap.tsv", name=None)
        output_figure = self._safe_plot(overlap_df=overlap_df)

        return ReportResult(
            name=self.name,
            output_figures=[output_figure] if output_figure is not None else [],
            output_tables=[output_table],
        )

    def _read_groundtruth_motifs(self, filepath):
        with open(filepath) as file:
            PositionalMotifHelper.check_file_header(file.readline(), filepath, expected_header="indices\tamino_acids\tn_sequences\n")
            groundtruth_motifs = []
            groundtruth_implant_rate = []
            for line in file.readlines():
                motif, implant_rate = self._get_motif_and_implant_rate(
                    line, motif_sep="\t"
                )
                groundtruth_motifs.append(motif)
                groundtruth_implant_rate.append(implant_rate)

        implant_rate_dict = {
            groundtruth_motifs[i]: groundtruth_implant_rate[i]
            for i in range(len(groundtruth_motifs))
        }
        return groundtruth_motifs, implant_rate_dict

    def _get_motif_and_implant_rate(self, string, motif_sep):
        indices_str, amino_acids_str, implant_rate = string.strip().split(motif_sep)
        motif = indices_str + "-" + amino_acids_str
        return motif, implant_rate

    def _generate_overlap(self, learned_motifs, groundtruth_motifs, implant_rate_dict):
        motif_size_list = list()
        implant_rate_list = list()
        max_overlap_list = list()
        learned_motif_list = list()
        gt_motif_list = list()

        for learned_motif in learned_motifs:
            motif_size = len(learned_motif.split("-")[0].replace("&", ""))
            for groundtruth_motif in groundtruth_motifs:
                max_overlap = self._get_max_overlap(learned_motif, groundtruth_motif)

                if max_overlap != 0:
                    motif_size_list.append(motif_size)
                    implant_rate_list.append(implant_rate_dict[groundtruth_motif])
                    max_overlap_list.append(max_overlap)
                    learned_motif_list.append(learned_motif)
                    gt_motif_list.append(groundtruth_motif)

        df = pd.DataFrame()
        df["learned_motif"] = learned_motif_list
        df["ground_truth_motif"] = gt_motif_list
        df["implant_rate"] = implant_rate_list
        df["max_overlap"] = max_overlap_list
        df["motif_size"] = motif_size_list

        return df

    def _get_max_overlap(self, learned_motif, groundtruth_motif):
        # assumes no duplicates will occur as is the case with motifs

        split_learned = learned_motif.replace("&", "").split("-")
        split_groundtruth = groundtruth_motif.replace("&", "").split("-")

        learned_aa = split_learned[0]
        learned_indices = split_learned[1]
        groundtruth_aa = split_groundtruth[0]
        groundtruth_indices = split_groundtruth[1]

        learned_pairs = [learned_aa[i] + learned_indices[i] for i in range(len(learned_aa))]
        groundtruth_pairs = [groundtruth_aa[i] + groundtruth_indices[i] for i in range(len(groundtruth_aa))]

        score = 0
        for pair in learned_pairs:
            if pair in groundtruth_pairs:
                score += 1

        return score

    def _get_color_discrete_sequence(self):
        return px.colors.qualitative.Pastel[:-1] + px.colors.qualitative.Set3

    def _plot(self, overlap_df) -> ReportOutput:
        file_path = self.result_path / f"motif_overlap.html"
        facet_barplot = px.histogram(
            overlap_df,
            x="implant_rate",
            labels={
                "implant_rate": "Number of implanted ground truth motifs",
                "max_overlap": "Ground truth motif overlap",
                "motif_size": "Motif size",
            },
            facet_col="max_overlap",
            color_discrete_sequence=self._get_color_discrete_sequence(),
            category_orders=dict(implant_rate=sorted([int(rate) for rate in overlap_df["implant_rate"].unique()]),
                                 motif_size=sorted([int(size) for size in overlap_df["motif_size"].unique()]),
                                 max_overlap=sorted([int(overlap) for overlap in overlap_df["max_overlap"].unique()])),
            facet_col_spacing=0.05,
            color="motif_size",
            title="Amount of overlapping motifs per implant rate",
            template="plotly_white"
        )
        facet_barplot.update_yaxes(matches=None, showticklabels=True)
        facet_barplot.update_layout(
            yaxis_title="Number of overlapping learned motifs",
        )
        facet_barplot.write_html(str(file_path))

        return ReportOutput(
            path=file_path, name="Amount of overlapping motifs per implant rate"
        )

    def check_prerequisites(self):
        valid_encodings = [MotifEncoder.__name__]

        if self.dataset.encoded_data is None or self.dataset.encoded_data.info is None:
            logging.warning(
                "GroundTruthMotifOverlap: the dataset is not encoded, skipping this report..."
            )
            return False
        elif self.dataset.encoded_data.encoding not in valid_encodings:
            logging.warning(
                f"GroundTruthMotifOverlap: the dataset encoding ({self.dataset.encoded_data.encoding}) was not in the list of valid "
                f"encodings ({valid_encodings}), skipping this report..."
            )
            return False
        else:
            return True



import os
from pathlib import Path

import numpy as np
import pandas as pd
from immuneML.ml_metrics.MetricUtil import MetricUtil

from immuneML.environment.Label import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.ml_metrics.MetricUtil import MetricUtil
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.steps.MLMethodAssessmentParams import MLMethodAssessmentParams
from immuneML.workflows.steps.Step import Step


class MLMethodAssessment(Step):
    fieldnames = ["run", "optimal_method_params", "method", "encoding_params", "encoding", "evaluated_on"]

    @staticmethod
    def run(input_params: MLMethodAssessmentParams = None):
        X = input_params.dataset.encoded_data
        predicted_y = input_params.method.predict(X, input_params.label)
        predicted_proba_y_per_class = input_params.method.predict_proba(X, input_params.label)
        true_y = input_params.dataset.encoded_data.labels
        example_weights = input_params.dataset.get_example_weights()

        example_ids = input_params.dataset.get_example_ids()

        MLMethodAssessment._store_predictions(method=input_params.method, true_y=true_y, predicted_y=predicted_y,
                                              predicted_proba_y_per_class=predicted_proba_y_per_class,
                                              label=input_params.label, predictions_path=input_params.predictions_path,
                                              example_ids=example_ids, split_index=input_params.split_index)

        scores = MLMethodAssessment._score(metrics_list=input_params.metrics, optimization_metric=input_params.optimization_metric,
                                           label=input_params.label, split_index=input_params.split_index, predicted_y=predicted_y,
                                           predicted_proba_y_per_class=predicted_proba_y_per_class, true_y=true_y,
                                           example_weights=example_weights, method=input_params.method,
                                           ml_score_path=input_params.ml_score_path)

        return scores

    @staticmethod
    def _score(metrics_list: set, optimization_metric: ClassificationMetric, label: Label, predicted_y, predicted_proba_y_per_class, true_y, example_weights, ml_score_path: Path,
               split_index: int, method: MLMethod):
        results = {}
        scores = {}

        predicted_proba_class = predicted_proba_y_per_class[label.name] if predicted_proba_y_per_class is not None else None
        predicted_proba_y = np.vstack([predicted_proba_class[cls] for cls in label.values]).T if predicted_proba_class is not None else None

        metrics_with_optim_metric = set(metrics_list)
        metrics_with_optim_metric.add(optimization_metric)

        metrics_with_optim_metric = sorted(list(metrics_with_optim_metric), key=lambda metric: metric.name)

        for metric in metrics_with_optim_metric:
            score = MetricUtil.score_for_metric(metric=metric,
                                                predicted_y=predicted_y[label.name],
                                                true_y=true_y[label.name],
                                                example_weights=example_weights,
                                                classes=label.values,
                                                predicted_proba_y=predicted_proba_y)
            results[f"{label.name}_{metric.name.lower()}"] = score
            scores[metric.name.lower()] = score

        results["split_index"] = split_index

        df = pd.DataFrame([results])

        if ml_score_path.is_file() and os.path.getsize(ml_score_path) > 0:
            df.to_csv(ml_score_path, mode='a', header=False, index=False)
        else:
            df.to_csv(ml_score_path, index=False)

        return scores

    @staticmethod
    def _store_predictions(method: MLMethod, true_y, predicted_y, predicted_proba_y_per_class, label: Label, predictions_path, summary_path=None,
                           example_ids: list = None, split_index: int = None):

        df = pd.DataFrame()
        df["example_id"] = example_ids
        df["split_index"] = [split_index for i in range(len(example_ids))]

        df[f"{label.name}_true_class"] = true_y[label.name]
        df[f"{label.name}_predicted_class"] = predicted_y[label.name]

        for cls in method.get_classes():
            tmp = predicted_proba_y_per_class[label.name][cls] if predicted_proba_y_per_class is not None and predicted_proba_y_per_class[label.name] is not None else None
            df[f'{label.name}_{cls}_proba'] = tmp

        if predictions_path is not None:
            df.to_csv(predictions_path, index=False)

        if summary_path is not None:
            PathBuilder.build(os.path.dirname(os.path.abspath(summary_path)))
            if os.path.isfile(summary_path) and os.path.getsize(summary_path) > 0:
                df.to_csv(summary_path, mode='a', header=False, index=False)
            else:
                df.to_csv(summary_path, index=False)



# quality: gold
from typing import Union, List

from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.data_model.receptor.receptor_sequence.SequenceFrameType import SequenceFrameType


class SequenceMetadata:
    """
    class modeling the existing knowledge about a receptor_sequence, should be stored according to
    AIRR nomenclature
        - v call
        - j call
        - chain
        - duplicate_count
        - region_type (e.g. IMGT_CDR3, IMGT_CDR1, FULL_SEQUENCE)
        - frame_type (e.g. IN, OUT, STOP)
        - sample
        - custom params (dictionary with custom sequence information)

    """

    def __init__(self, v_call: str = None, j_call: str = None, chain=None, duplicate_count: int = None,
                 frame_type: Union[SequenceFrameType, str] = '',
                 region_type: str = None, cell_id: str = None, custom_params: dict = None):
        self.v_call = v_call
        self.j_call = j_call
        self.chain = Chain.get_chain(chain) if chain and isinstance(chain, str) else chain if isinstance(chain,
                                                                                                         Chain) else None
        self.duplicate_count = int(float(duplicate_count)) if isinstance(duplicate_count, str) else duplicate_count
        self.frame_type = SequenceFrameType(frame_type.upper()) \
            if frame_type and isinstance(frame_type, str) and frame_type != 'nan' else frame_type if isinstance(
            frame_type, SequenceFrameType) else SequenceFrameType.UNDEFINED
        self.region_type = RegionType(region_type.upper()) \
            if region_type and isinstance(region_type, str) and region_type != 'nan' else region_type if isinstance(
            region_type, RegionType) else None
        self.cell_id = cell_id
        self.custom_params = custom_params if custom_params is not None else {}

    @property
    def v_gene(self):
        return self.v_call.split("\*")[0]

    @property
    def j_gene(self):
        return self.j_call.split("\*")[0]

    def get_attribute(self, name: str):
        """Returns the attribute value if attribute is present either directly or in custom_params, otherwise returns
        None"""
        if hasattr(self, name):
            return getattr(self, name)
        elif name in self.custom_params:
            return self.custom_params[name]
        else:
            return None

    def get_all_attribute_names(self) -> List[str]:
        return [el for el in vars(self) if el != 'custom_params'] + list(self.custom_params.keys())



import plotly.express as px
import warnings
from pathlib import Path

import pandas as pd

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder
from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper

class WeightsDistribution(DataReport):
    """
    Plots the distribution of weights in a given Dataset. This report can only be used if example weighting has been applied to the given dataset.


    # todo: the report should work with any label, with any number of classes (currently assumes is_binder with classes 0 and 1)
    # use self.label property to find the label name.
    # do not hardcode any text related to the is_binder label

    # todo: make sure the same classes always get the same color in the figures
    # currently when running the report multiple times, class 0 is sometimes blue and sometimes purple
    # solution: retrieve the classes, sort the classes, use color_discrete_map instead of color_discrete_sequence
    # to map each class to a color. Get these colors for example from px.colors.diverging.Tealrose

    # todo: label is only a useful parameter when split_classes is true. these parameters can thus be merged
    # instead, use parameter color_grouping_label. if set, use different colors. if not set, use one color
    # see FeatureComparison report for an example where multiple labels can be used to change features of the plot

    # todo add unit tests
    # todo add the sequence as one of the hover values (plotly argument hover_data=["..."]), see example: https://plotly.com/python/hover-text-and-formatting/


    Example YAML specification:
        r1:
      WeightsDistribution:
        label:
          is_binding
        weight_thresholds:
        - 1
        - 0.1
        - 0.001
        split_classes:
          True
    """
    @classmethod
    def build_object(cls, **kwargs):
        return WeightsDistribution(**kwargs)

    def __init__(self, dataset: Dataset = None, result_path: Path = None, number_of_processes: int = 1, name: str = None, label: dict = None, weight_thresholds: dict = None, split_classes: bool = None):
            super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
            self.label = label
            self.weight_thresholds = weight_thresholds
            self.split_classes = split_classes
            self.label_config = None

    def check_prerequisites(self):
        if self.dataset.get_example_weights() is not None:
            return True
        else:
            warnings.warn("WeightsDistribution: report requires weights to be set for the given Dataset. Skipping this report...")
            return False

    def _generate(self) -> ReportResult:
        self.label_config = LabelHelper.create_label_config([self.label], self.dataset, WeightsDistribution.__name__,
                                                       f"{WeightsDistribution.__name__}/label")
        data = self._get_plotting_data()
        report_output_fig = self._safe_plot(data=data)
        output_figures = None if report_output_fig is None else [report_output_fig]
        return ReportResult(name=self.name,
                                info="A line graph of weights over sequences for is-binding and non-binding sequences",
                                output_figures=output_figures)


    def get_weights_by_class_df(self, data, weights):
        data["weights"] = weights
        data.sort_values(by=["weights"], inplace=True, ascending=False)

        data["sorted"] = None
        data.loc[data["is_binding"] == "0", "sorted"] = range(len(data[data["is_binding"] == "0"]))
        data.loc[data["is_binding"] == "1", "sorted"] = range(len(data[data["is_binding"] == "1"]))

        return data

    def get_weights_both_classes_df(self, data, weights):
        data["weights"] = weights
        data.sort_values(by=["weights"], inplace=True, ascending=False)
        data.loc[:, "sorted"] = range(len(data))

        return data

    def _get_plotting_data(self):
        weights = self.dataset.get_example_weights()
        data = self.dataset.get_metadata([self.label], return_df=True)

        # todo remove this, this was temporary
        if isinstance(self.dataset, SequenceDataset):
            data["seq"] = [seq.sequence_aa for seq in self.dataset.get_data()]

        if self.split_classes:
            data = self.get_weights_by_class_df(data, weights)
        else:
            data = self.get_weights_both_classes_df(data, weights)

        return data

    def _plot(self, data) -> ReportOutput:
        print(data)
        PathBuilder.build(self.result_path)

        if self.split_classes:
            color = self.label
        else:
            color = None

        fig = px.line(data, y="weights", color=color, x="sorted", log_y=True,
                      color_discrete_sequence=["#65D0B8", "#AB80D8"],
                      hover_data=["seq"],
                      labels={
                          "weights": "Weights",
                          "sorted": "Sequences (sorted by weight)",
                          "is_binding": "Is binder"
                      }, template="plotly_white")

        if self.weight_thresholds is not None:
            for hline in self.weight_thresholds:
                fig.add_shape(type="line", x0=0, x1=max(data["sorted"]), y0=hline, y1=hline, line=dict(color="black", width=1.5, dash="dash"))

        file_path = self.result_path / "weights_distribution.html"
        fig.write_html(str(file_path))
        return ReportOutput(path=file_path, name="weights distribution plot")


# quality: gold

import abc

from immuneML.data_model.dataset.Dataset import Dataset


class DataImport(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def import_dataset(params, dataset_name: str) -> Dataset:
        pass




from pathlib import Path

from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.workflows.instructions.GenModelInstruction import GenModelInstruction
from immuneML.workflows.instructions.GenModelInstruction import GenModelState


class ApplyGenModelState(GenModelState):
    pass


class ApplyGenModelInstruction(GenModelInstruction):
    """
    ApplyGenModel instruction implements applying generative AIRR models on the sequence level.

    This instruction takes as input a trained model which will be used for generating data and the number of
    sequences to be generated. It can also produce reports of the applied model and reports of generated
    sequences.

    To train generative model with immuneML, see TrainGenModel instruction.

    .. note::

        This is an experimental feature in version 3.0.0a1.

    Specification arguments:

    - gen_examples_count (int): how many examples (sequences, repertoires) to generate from the applied model

    - reports (list): list of report ids (defined under definitions/reports) to apply after generating
      gen_examples_count examples; these can be data reports (to be run on generated examples), ML reports (to be run
      on the fitted model)

    - config_path (str): path to the trained model in zip format (as provided by TrainGenModel instruction)

    YAML specification:

    .. highlight:: yaml
    .. code-block:: yaml

        my_apply_gen_model_inst: # user-defined instruction name
            type: ApplyGenModel
            gen_examples_count: 100
            ml_config_path: ./config.zip
            reports: [data_rep1, ml_rep2]

    """
    def __init__(self, method: GenerativeModel = None, reports: list = None, result_path: Path = None,
                 name: str = None, gen_examples_count: int = None):
        super().__init__(ApplyGenModelState(result_path, name, gen_examples_count), method, reports)

    def run(self, result_path: Path) -> GenModelState:
        self._set_path(result_path)
        self._gen_data()
        self._run_reports()

        return self.state



from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.simulation.dataset_generation.RandomDatasetGenerator import RandomDatasetGenerator
from immuneML.util.ParameterValidator import ParameterValidator


class RandomRepertoireDatasetImport(DataImport):
    """
    Returns a RepertoireDataset consisting of randomly generated sequences, which can be used for benchmarking purposes.
    The sequences consist of uniformly chosen amino acids or nucleotides.

    Specification arguments:

    - repertoire_count (int): The number of repertoires the RepertoireDataset should contain.

    - sequence_count_probabilities (dict): A mapping where the keys are the number of sequences per repertoire, and the values are the probabilities that any of the repertoires would have that number of sequences. For example, to create a random RepertoireDataset where 40% of the repertoires would have 1000 sequences, and the other 60% would have 1100 sequences, this mapping would need to be specified:

        .. indent with spaces
        .. code-block:: yaml

                1000: 0.4
                1100: 0.6

    - sequence_length_probabilities (dict): A mapping where the keys correspond to different sequence lengths, and the values are the probabilities for choosing each sequence length. For example, to create a random RepertoireDataset where 40% of the sequences would be of length 10, and 60% of the sequences would have length 12, this mapping would need to be specified:

        .. indent with spaces
        .. code-block:: yaml

                10: 0.4
                12: 0.6

    - labels (dict): A mapping that specifies randomly chosen labels to be assigned to the Repertoires. One or multiple labels can be specified here. The keys of this mapping are the labels, and the values consist of another mapping between label classes and their probabilities. For example, to create a random RepertoireDataset with the label CMV where 70% of the Repertoires has class cmv_positive and the remaining 30% has class cmv_negative, the following mapping should be specified:

        .. indent with spaces
        .. code-block:: yaml

                CMV:
                    cmv_positive: 0.7
                    cmv_negative: 0.3


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_random_dataset:
            format: RandomRepertoireDataset
            params:
                repertoire_count: 100 # number of random repertoires to generate
                sequence_count_probabilities:
                    10: 0.5 # probability that any of the repertoires would have 10 receptor sequences
                    20: 0.5
                sequence_length_probabilities:
                    10: 0.5 # probability that any of the receptor sequences would be 10 amino acids in length
                    12: 0.5
                labels: # randomly assigned labels (only useful for simple benchmarking)
                    cmv:
                        True: 0.5 # probability of value True for label cmv to be assigned to any repertoire
                        False: 0.5

    """

    @staticmethod
    def import_dataset(params: dict, dataset_name: str) -> RepertoireDataset:
        valid_keys = ["result_path", "repertoire_count", "sequence_count_probabilities", "sequence_length_probabilities", "labels"]
        ParameterValidator.assert_all_in_valid_list(list(params.keys()), valid_keys, "RandomRepertoireDatasetImport", "params")

        return RandomDatasetGenerator.generate_repertoire_dataset(repertoire_count=params["repertoire_count"],
                                                                  sequence_count_probabilities=params["sequence_count_probabilities"],
                                                                  sequence_length_probabilities=params["sequence_length_probabilities"],
                                                                  labels=params["labels"],
                                                                  path=params["result_path"])



import uuid
from pathlib import Path

import pandas as pd

from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.data_model.receptor.receptor_sequence.SequenceMetadata import SequenceMetadata
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.util.PathBuilder import PathBuilder


class RepertoireBuilder:
    """
    Helper class for tests: creates repertoires from a list of a list of sequences and stores them in the given path
    """
    @staticmethod
    def build(sequences: list, path: Path, labels: dict = None, seq_metadata: list = None, subject_ids: list = None):

        if subject_ids is not None:
            assert len(subject_ids) == len(sequences)

        if seq_metadata is not None:
            assert len(sequences) == len(seq_metadata)
            for index, sequence_list in enumerate(sequences):
                assert len(sequence_list) == len(seq_metadata[index])

        assert not (path / "repertoires").is_dir(), f"RepertoireBuilder: attempted to store new repertoires at {path / 'repertoires'} but this folder already exists. " \
                                                    f"Please remove this folder or specify a different path. "

        PathBuilder.build(path)
        rep_path = PathBuilder.build(path / "repertoires")

        repertoires = []
        if subject_ids is None:
            subject_ids = []

        for rep_index, sequence_list in enumerate(sequences):
            rep_sequences = []
            if len(subject_ids) < len(sequences):
                subject_ids.append("rep_" + str(rep_index))
            for seq_index, sequence in enumerate(sequence_list):
                if seq_metadata is None:
                    m = SequenceMetadata(v_call="TRBV1-1*01", j_call="TRBJ1-1*01", duplicate_count=1, chain="TRB", region_type="IMGT_CDR3")
                else:
                    m = SequenceMetadata(**seq_metadata[rep_index][seq_index])

                s = ReceptorSequence(sequence_aa=sequence, metadata=m, sequence_id=str(uuid.uuid4().hex))
                rep_sequences.append(s)

            if labels is not None:
                metadata = {key: labels[key][rep_index] for key in labels.keys()}
            else:
                metadata = {}

            metadata = {**metadata, **{"subject_id": subject_ids[rep_index]}}

            repertoire = Repertoire.build_from_sequence_objects(rep_sequences, rep_path, metadata, filename_base=f"rep_{rep_index}")
            repertoires.append(repertoire)

        df = pd.DataFrame({**{"filename": [repertoire.data_filename for repertoire in repertoires],
                              "subject_id": subject_ids,
                              "identifier": [repertoire.identifier for repertoire in repertoires]},
                           **(labels if labels is not None else {})})
        df.to_csv(path / "metadata.csv", index=False)

        return repertoires, path / "metadata.csv"
